import dolfin
import ufl_legacy as ufl
from pyadjoint import Block, OverloadedType, no_annotations

from fenics_adjoint.utils import function_from_vector, extract_subfunction


def create_bc(bc, value=None, homogenize=None):
    """Create a new bc object from an existing one.

    :arg bc: The :class:`~.DirichletBC` to clone.
    :arg value: A new value to use.
    :arg homogenize: If True, return a homogeneous version of the bc.

    One cannot provide both ``value`` and ``homogenize``, but
    should provide at least one.
    """
    if value is None and homogenize is None:
        raise ValueError("No point cloning a bc if you're not changing its values")
    if value is not None and homogenize is not None:
        raise ValueError("Cannot provide both value and homogenize")
    if homogenize:
        bc = dolfin.DirichletBC(bc)
        bc.homogenize()
        return bc
    try:
        # FIXME: Not perfect handling of Initialization, wait for development in dolfin.DirichletBC
        bc = dolfin.DirichletBC(dolfin.FunctionSpace(bc.function_space()),
                                value, *bc.domain_args)
    except AttributeError:
        from IPython import embed

        bc = dolfin.DirichletBC(dolfin.FunctionSpace(bc.function_space()),
                                value,
                                bc.sub_domain, method=bc.method())
    return bc


def extract_bc_subvector(value, Vtarget, bc):
    """Extract from value (a function in a mixed space), the sub
    function corresponding to the part of the space bc applies
    to.  Vtarget is the target (collapsed) space."""
    assigner = dolfin.FunctionAssigner(Vtarget, dolfin.FunctionSpace(bc.function_space()))
    output = dolfin.Function(Vtarget)
    assigner.assign(output, extract_subfunction(value, dolfin.FunctionSpace(bc.function_space())))
    return output.vector()


class DirichletBCBlock(Block):
    def __init__(self, *args, **kwargs):
        Block.__init__(self, ad_block_tag=kwargs.pop('ad_block_tag', None))
        self.function_space = args[0]
        self.parent_space = self.function_space
        while hasattr(self.parent_space, "_ad_parent_space") and self.parent_space._ad_parent_space is not None:
            self.parent_space = self.parent_space._ad_parent_space
        self.collapsed_space = self.function_space
        if self.function_space != self.parent_space:
            self.collapsed_space = self.function_space.collapse()

        if len(args) >= 2 and isinstance(args[1], OverloadedType):
            self.add_dependency(args[1])
        else:
            # TODO: Implement the other cases.
            #       Probably just a BC without dependencies?
            #       In which case we might not even need this Block?
            # Update: What if someone runs: `DirichletBC(V, g*g, "on_boundary")`.
            #         In this case the dolfin will project the product onto V.
            #         But we will have to annotate the product somehow.
            #         One solution would be to do a check and add a ProjectBlock before the DirichletBCBlock.
            #         (Either by actually running our project or by "manually" inserting a project block).
            pass

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        bc = self.get_outputs()[0].saved_output
        c = block_variable.output
        adj_inputs = adj_inputs[0]
        adj_output = None
        for adj_input in adj_inputs:
            if isinstance(c, dolfin.Constant):
                adj_value = dolfin.Function(self.parent_space)
                adj_input.apply(adj_value.vector())
                if self.function_space != self.parent_space:
                    vec = extract_bc_subvector(adj_value, self.collapsed_space, bc)
                    adj_value = function_from_vector(self.collapsed_space, vec)

                if adj_value.ufl_shape == () or adj_value.ufl_shape[0] <= 1:
                    r = adj_value.vector().sum()
                else:
                    output = []
                    subindices = _extract_subindices(self.function_space)
                    for indices in subindices:
                        current_subfunc = adj_value
                        prev_idx = None
                        for i in indices:
                            if prev_idx is not None:
                                current_subfunc = current_subfunc.sub(prev_idx)
                            prev_idx = i
                        output.append(current_subfunc.sub(prev_idx, deepcopy=True).vector().sum())

                    r = dolfin.cpp.la.Vector(dolfin.MPI.comm_world, len(output))
                    r[:] = output
            elif isinstance(c, dolfin.Function):
                # TODO: This gets a little complicated.
                #       The function may belong to a different space,
                #       and with `Function.set_allow_extrapolation(True)`
                #       you can even use the Function outside its domain.
                # For now we will just assume the FunctionSpace is the same for
                # the BC and the Function.
                adj_value = dolfin.Function(self.parent_space)
                adj_input.apply(adj_value.vector())
                r = extract_bc_subvector(adj_value, c.function_space(), bc)
            elif isinstance(c, dolfin.Expression):
                adj_value = dolfin.Function(self.parent_space)
                adj_input.apply(adj_value.vector())
                output = extract_bc_subvector(adj_value, self.collapsed_space, bc)
                r = [[output, self.collapsed_space]]
            if adj_output is None:
                adj_output = r
            else:
                adj_output += r
        return adj_output

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        bc = block_variable.saved_output
        for bv in self.get_dependencies():
            tlm_input = bv.tlm_value

            if tlm_input is None:
                continue

            if self.function_space != self.parent_space and not isinstance(tlm_input, ufl.Coefficient):
                tlm_input = dolfin.project(tlm_input, self.collapsed_space)

            # TODO: This is gonna crash for dirichletbcs with multiple dependencies (can't add two bcs)
            #       However, if there is multiple dependencies, we need to AD the expression (i.e if value=f*g then
            #       dvalue = tlm_f * g + f * tlm_g). Right now we can only handle value=f => dvalue = tlm_f.
            m = create_bc(bc, value=tlm_input)
        return m

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        # The same as evaluate_adj but with hessian values.
        return self.evaluate_adj_component(inputs, hessian_inputs, block_variable, idx)

    @no_annotations
    def recompute(self):
        # There is nothing to do. The checkpoint is weak,
        # so it changes automatically with the dependency checkpoint.
        return

    def __str__(self):
        return "DirichletBC block"


def _extract_subindices(V):
    assert V.num_sub_spaces() > 0
    r = []
    for i in range(V.num_sub_spaces()):
        indices_sequence = [i]
        _build_subindices(indices_sequence, r, V.sub(i))
        indices_sequence.pop()
    return r


def _build_subindices(indices_sequence, r, V):
    if V.num_sub_spaces() <= 0:
        r.append(tuple(indices_sequence))
    else:
        for i in range(V.num_sub_spaces()):
            indices_sequence.append(i)
            _build_subindices(indices_sequence, r, V)
            indices_sequence.pop()



import dolfin

__all__ = []
backend_genericmatrix_mul = dolfin.cpp.la.GenericMatrix.__mul__


def adjoint_genericmatrix_mul(self, other):
    out = backend_genericmatrix_mul(self, other)
    if hasattr(self, 'form') and isinstance(other, dolfin.cpp.la.GenericVector):
        if hasattr(other, 'form'):
            out.form = dolfin.action(self.form, other.form)
        elif hasattr(other, 'function'):
            if hasattr(other, 'function_factor'):
                out.form = dolfin.action(other.function_factor * self.form, other.function)
            else:
                out.form = dolfin.action(self.form, other.function)

    return out


dolfin.cpp.la.GenericMatrix.__mul__ = adjoint_genericmatrix_mul

backend_genericmatrix_ident_zeros = dolfin.cpp.la.GenericMatrix.ident_zeros


def ident_zeros(self, tol=dolfin.DOLFIN_EPS):
    backend_genericmatrix_ident_zeros(self, tol)
    self.ident_zeros_tol = tol


dolfin.cpp.la.GenericMatrix.ident_zeros = ident_zeros



import dolfin
import numpy
import ufl_legacy as ufl
from pyadjoint import Block
from pyadjoint.enlisting import Enlist
from ufl_legacy.formatting.ufl2unicode import ufl2unicode

from fenics_adjoint.utils import function_from_vector, extract_subfunction, create_function, extract_mesh_from_form, linalg_solve
from .assembly import assemble_adjoint_value
from .dirichlet_bc import create_bc


class GenericSolveBlock(Block):
    pop_kwargs_keys = ["adj_cb", "adj_bdy_cb", "adj2_cb", "adj2_bdy_cb",
                       "forward_args", "forward_kwargs", "adj_args", "adj_kwargs"]

    def __init__(self, lhs, rhs, func, bcs, *args, **kwargs):
        super().__init__(ad_block_tag=kwargs.pop('ad_block_tag', None))
        self.adj_cb = kwargs.pop("adj_cb", None)
        self.adj_bdy_cb = kwargs.pop("adj_bdy_cb", None)
        self.adj2_cb = kwargs.pop("adj2_cb", None)
        self.adj2_bdy_cb = kwargs.pop("adj2_bdy_cb", None)
        self.adj_sol = None

        self.forward_args = []
        self.forward_kwargs = {}
        self.adj_args = []
        self.adj_kwargs = {}
        self.assemble_kwargs = {}

        # Equation LHS
        self.lhs = lhs
        # Equation RHS
        self.rhs = rhs
        # Solution function
        self.func = func
        self.function_space = self.func.function_space()
        # Boundary conditions
        self.bcs = []
        if bcs is not None:
            self.bcs = Enlist(bcs)

        if isinstance(self.lhs, ufl.Form) and isinstance(self.rhs, ufl.Form):
            self.linear = True
            for c in self.rhs.coefficients():
                self.add_dependency(c, no_duplicates=True)
        else:
            self.linear = False

        for c in self.lhs.coefficients():
            self.add_dependency(c, no_duplicates=True)

        for bc in self.bcs:
            self.add_dependency(bc, no_duplicates=True)

        mesh = self.lhs.ufl_domain().ufl_cargo()
        self.add_dependency(mesh)
        self._init_solver_parameters(args, kwargs)

    def _init_solver_parameters(self, args, kwargs):
        self.forward_args = kwargs.pop("forward_args", [])
        self.forward_kwargs = kwargs.pop("forward_kwargs", {})
        self.adj_args = kwargs.pop("adj_args", [])
        self.adj_kwargs = kwargs.pop("adj_kwargs", {})
        self.assemble_kwargs = {}

    def __str__(self):
        return "solve({} = {})".format(ufl2unicode(self.lhs),
                                       ufl2unicode(self.rhs))

    def _create_F_form(self):
        # Process the equation forms, replacing values with checkpoints,
        # and gathering lhs and rhs in one single form.
        if self.linear:
            tmp_u = create_function(self.function_space)
            F_form = dolfin.action(self.lhs, tmp_u) - self.rhs
        else:
            tmp_u = self.func
            F_form = self.lhs

        replace_map = self._replace_map(F_form)
        replace_map[tmp_u] = self.get_outputs()[0].saved_output
        return ufl.replace(F_form, replace_map)

    def _homogenize_bcs(self):
        bcs = []
        for bc in self.bcs:
            if isinstance(bc, dolfin.DirichletBC):
                bc = create_bc(bc, homogenize=True)
            bcs.append(bc)
        return bcs

    def _create_initial_guess(self):
        return dolfin.Function(self.function_space)

    def _recover_bcs(self):
        bcs = []
        for block_variable in self.get_dependencies():
            c = block_variable.output
            c_rep = block_variable.saved_output

            if isinstance(c, dolfin.DirichletBC):
                bcs.append(c_rep)
        return bcs

    def _replace_map(self, form):
        replace_coeffs = {}
        for block_variable in self.get_dependencies():
            coeff = block_variable.output
            if coeff in form.coefficients():
                replace_coeffs[coeff] = block_variable.saved_output
        return replace_coeffs

    def _replace_form(self, form, func=None):
        """Replace the form coefficients with checkpointed values

        func represents the initial guess if relevant.
        """
        replace_map = self._replace_map(form)
        if func is not None and self.func in replace_map:
            dolfin.Function.assign(func, replace_map[self.func])
            replace_map[self.func] = func
        return ufl.replace(form, replace_map)

    def _should_compute_boundary_adjoint(self, relevant_dependencies):
        # Check if DirichletBC derivative is relevant
        bdy = False
        for _, dep in relevant_dependencies:
            if isinstance(dep.output, dolfin.DirichletBC):
                bdy = True
                break
        return bdy

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        fwd_block_variable = self.get_outputs()[0]
        u = fwd_block_variable.output

        dJdu = adj_inputs[0]

        F_form = self._create_F_form()

        dFdu = dolfin.derivative(F_form,
                                 fwd_block_variable.saved_output,
                                 dolfin.TrialFunction(u.function_space()))
        dFdu_form = dolfin.adjoint(dFdu)
        dJdu = dJdu.copy()

        compute_bdy = self._should_compute_boundary_adjoint(relevant_dependencies)
        adj_sol, adj_sol_bdy = self._assemble_and_solve_adj_eq(dFdu_form, dJdu, compute_bdy)
        self.adj_sol = adj_sol
        if self.adj_cb is not None:
            self.adj_cb(adj_sol)
        if self.adj_bdy_cb is not None and compute_bdy:
            self.adj_bdy_cb(adj_sol_bdy)

        r = {}
        r["form"] = F_form
        r["adj_sol"] = adj_sol
        r["adj_sol_bdy"] = adj_sol_bdy
        return r

    def _assemble_and_solve_adj_eq(self, dFdu_adj_form, dJdu, compute_bdy):
        dJdu_copy = dJdu.copy()
        kwargs = self.assemble_kwargs.copy()
        # Homogenize and apply boundary conditions on adj_dFdu and dJdu.
        bcs = self._homogenize_bcs()
        kwargs["bcs"] = bcs
        dFdu = assemble_adjoint_value(dFdu_adj_form, **kwargs)

        for bc in bcs:
            bc.apply(dJdu)

        adj_sol = create_function(self.function_space)
        linalg_solve(dFdu, adj_sol.vector(), dJdu, *self.adj_args, **self.adj_kwargs)

        adj_sol_bdy = None
        if compute_bdy:
            adj_sol_bdy = function_from_vector(self.function_space,
                                               dJdu_copy - assemble_adjoint_value(
                                                   dolfin.action(dFdu_adj_form, adj_sol)))

        return adj_sol, adj_sol_bdy

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        if not self.linear and self.func == block_variable.output:
            # We are not able to calculate derivatives wrt initial guess.
            return None
        F_form = prepared["form"]
        adj_sol = prepared["adj_sol"]
        adj_sol_bdy = prepared["adj_sol_bdy"]
        c = block_variable.output
        c_rep = block_variable.saved_output

        if isinstance(c, dolfin.Constant):
            mesh = extract_mesh_from_form(F_form)
            trial_function = dolfin.TrialFunction(c._ad_function_space(mesh))
        elif isinstance(c, dolfin.Function):
            trial_function = dolfin.TrialFunction(c.function_space())
        elif isinstance(c, dolfin.function.expression.BaseExpression):
            mesh = F_form.ufl_domain().ufl_cargo()
            c_fs = c._ad_function_space(mesh)
            trial_function = dolfin.TrialFunction(c_fs)
        elif isinstance(c, dolfin.DirichletBC):
            tmp_bc = create_bc(c, value=extract_subfunction(adj_sol_bdy, c.function_space()))
            return [tmp_bc]
        elif isinstance(c, dolfin.Mesh):
            # Using CoordianteDerivative requires us to do action before
            # differentiating, might change in the future.
            F_form_tmp = dolfin.action(F_form, adj_sol)
            X = dolfin.SpatialCoordinate(c_rep)
            dFdm = dolfin.derivative(-F_form_tmp, X, dolfin.TestFunction(c._ad_function_space()))

            dFdm = assemble_adjoint_value(dFdm, **self.assemble_kwargs)
            return dFdm

        dFdm = -dolfin.derivative(F_form, c_rep, trial_function)
        dFdm = dolfin.adjoint(dFdm)
        dFdm = dFdm * adj_sol
        dFdm = assemble_adjoint_value(dFdm, **self.assemble_kwargs)
        if isinstance(c, dolfin.function.expression.BaseExpression):
            return [[dFdm, c_fs]]
        else:
            return dFdm

    def prepare_evaluate_tlm(self, inputs, tlm_inputs, relevant_outputs):
        fwd_block_variable = self.get_outputs()[0]
        u = fwd_block_variable.output

        F_form = self._create_F_form()

        # Obtain dFdu.
        dFdu = dolfin.derivative(F_form,
                                 fwd_block_variable.saved_output,
                                 dolfin.TrialFunction(u.function_space()))

        return {
            "form": F_form,
            "dFdu": dFdu
        }

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        F_form = prepared["form"]
        dFdu = prepared["dFdu"]
        V = self.get_outputs()[idx].output.function_space()

        bcs = []
        dFdm = 0.
        for block_variable in self.get_dependencies():
            tlm_value = block_variable.tlm_value
            c = block_variable.output
            c_rep = block_variable.saved_output

            if isinstance(c, dolfin.DirichletBC):
                if tlm_value is None:
                    bcs.append(create_bc(c, homogenize=True))
                else:
                    bcs.append(tlm_value)
                continue
            elif isinstance(c, dolfin.Mesh):
                X = dolfin.SpatialCoordinate(c)
                c_rep = X

            if tlm_value is None:
                continue

            if c == self.func and not self.linear:
                continue

            dFdm += dolfin.derivative(-F_form, c_rep, tlm_value)

        if isinstance(dFdm, float):
            v = dFdu.arguments()[0]
            dFdm = dolfin.inner(dolfin.Constant(numpy.zeros(v.ufl_shape)), v) * dolfin.dx

        dFdm = ufl.algorithms.expand_derivatives(dFdm)
        dFdm = assemble_adjoint_value(dFdm)
        dudm = dolfin.Function(V)
        return self._assemble_and_solve_tlm_eq(
            assemble_adjoint_value(dFdu, bcs=bcs, **self.assemble_kwargs), dFdm, dudm, bcs)

    def _assemble_and_solve_tlm_eq(self, dFdu, dFdm, dudm, bcs):
        return self._assembled_solve(dFdu, dFdm, dudm, bcs)

    def _assemble_soa_eq_rhs(self, dFdu_form, adj_sol, hessian_input, d2Fdu2):
        # Start piecing together the rhs of the soa equation
        b = hessian_input.copy()
        if len(d2Fdu2.integrals()) > 0:
            b_form = dolfin.action(dolfin.adjoint(d2Fdu2), adj_sol)
        else:
            b_form = d2Fdu2

        for bo in self.get_dependencies():
            c = bo.output
            c_rep = bo.saved_output
            tlm_input = bo.tlm_value

            if (c == self.func and not self.linear) or tlm_input is None:
                continue

            if isinstance(c, dolfin.Mesh):
                X = dolfin.SpatialCoordinate(c)
                dFdu_adj = dolfin.action(dolfin.adjoint(dFdu_form), adj_sol)
                d2Fdudm = ufl.algorithms.expand_derivatives(
                    dolfin.derivative(dFdu_adj, X, tlm_input))
                if len(d2Fdudm.integrals()) > 0:
                    b_form += d2Fdudm
            elif not isinstance(c, dolfin.DirichletBC):
                dFdu_adj = dolfin.action(dolfin.adjoint(dFdu_form), adj_sol)
                b_form += dolfin.derivative(dFdu_adj, c_rep, tlm_input)

        b_form = ufl.algorithms.expand_derivatives(b_form)
        if len(b_form.integrals()) > 0:
            b -= assemble_adjoint_value(b_form)

        return b

    def _assemble_and_solve_soa_eq(self, dFdu_form, adj_sol, hessian_input, d2Fdu2, compute_bdy):
        b = self._assemble_soa_eq_rhs(dFdu_form, adj_sol, hessian_input, d2Fdu2)
        dFdu_form = dolfin.adjoint(dFdu_form)
        adj_sol2, adj_sol2_bdy = self._assemble_and_solve_adj_eq(dFdu_form, b, compute_bdy)
        if self.adj2_cb is not None:
            self.adj2_cb(adj_sol2)
        if self.adj2_bdy_cb is not None and compute_bdy:
            self.adj2_bdy_cb(adj_sol2_bdy)
        return adj_sol2, adj_sol2_bdy

    def prepare_evaluate_hessian(self, inputs, hessian_inputs, adj_inputs, relevant_dependencies):
        # First fetch all relevant values
        fwd_block_variable = self.get_outputs()[0]
        hessian_input = hessian_inputs[0]
        tlm_output = fwd_block_variable.tlm_value

        if hessian_input is None:
            return

        if tlm_output is None:
            return

        F_form = self._create_F_form()

        # Using the equation Form we derive dF/du, d^2F/du^2 * du/dm * direction.
        dFdu_form = dolfin.derivative(F_form, fwd_block_variable.saved_output)
        d2Fdu2 = ufl.algorithms.expand_derivatives(
            dolfin.derivative(dFdu_form, fwd_block_variable.saved_output, tlm_output))

        adj_sol = self.adj_sol
        if adj_sol is None:
            raise RuntimeError("Hessian computation was run before adjoint.")
        bdy = self._should_compute_boundary_adjoint(relevant_dependencies)
        adj_sol2, adj_sol2_bdy = self._assemble_and_solve_soa_eq(dFdu_form, adj_sol, hessian_input, d2Fdu2, bdy)

        r = {}
        r["adj_sol2"] = adj_sol2
        r["adj_sol2_bdy"] = adj_sol2_bdy
        r["form"] = F_form
        r["adj_sol"] = adj_sol
        return r

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        c = block_variable.output
        if c == self.func and not self.linear:
            return None

        adj_sol2 = prepared["adj_sol2"]
        adj_sol2_bdy = prepared["adj_sol2_bdy"]
        F_form = prepared["form"]
        adj_sol = prepared["adj_sol"]
        fwd_block_variable = self.get_outputs()[0]
        tlm_output = fwd_block_variable.tlm_value

        c_rep = block_variable.saved_output

        # If m = DirichletBC then d^2F(u,m)/dm^2 = 0 and d^2F(u,m)/dudm = 0,
        # so we only have the term dF(u,m)/dm * adj_sol2
        if isinstance(c, dolfin.DirichletBC):
            tmp_bc = create_bc(c, value=extract_subfunction(adj_sol2_bdy, c.function_space()))
            return [tmp_bc]

        if isinstance(c_rep, dolfin.Constant):
            mesh = extract_mesh_from_form(F_form)
            W = c._ad_function_space(mesh)
        elif isinstance(c, dolfin.function.expression.BaseExpression):
            mesh = F_form.ufl_domain().ufl_cargo()
            W = c._ad_function_space(mesh)
        elif isinstance(c, dolfin.Mesh):
            X = dolfin.SpatialCoordinate(c)
            W = c._ad_function_space()
        else:
            W = c.function_space()

        dc = dolfin.TestFunction(W)
        form_adj = dolfin.action(F_form, adj_sol)
        form_adj2 = dolfin.action(F_form, adj_sol2)
        if isinstance(c, dolfin.Mesh):
            dFdm_adj = dolfin.derivative(form_adj, X, dc)
            dFdm_adj2 = dolfin.derivative(form_adj2, X, dc)
        else:
            dFdm_adj = dolfin.derivative(form_adj, c_rep, dc)
            dFdm_adj2 = dolfin.derivative(form_adj2, c_rep, dc)

        # TODO: Old comment claims this might break on split. Confirm if true or not.
        d2Fdudm = ufl.algorithms.expand_derivatives(
            dolfin.derivative(dFdm_adj, fwd_block_variable.saved_output,
                              tlm_output))

        d2Fdm2 = 0
        # We need to add terms from every other dependency
        # i.e. the terms d^2F/dm_1dm_2
        for _, bv in relevant_dependencies:
            c2 = bv.output
            c2_rep = bv.saved_output

            if isinstance(c2, dolfin.DirichletBC):
                continue

            tlm_input = bv.tlm_value
            if tlm_input is None:
                continue

            if c2 == self.func and not self.linear:
                continue

            # TODO: If tlm_input is a Sum, this crashes in some instances?
            if isinstance(c2_rep, dolfin.Mesh):
                X = dolfin.SpatialCoordinate(c2_rep)
                d2Fdm2 += ufl.algorithms.expand_derivatives(dolfin.derivative(dFdm_adj, X, tlm_input))
            else:
                d2Fdm2 += ufl.algorithms.expand_derivatives(dolfin.derivative(dFdm_adj, c2_rep, tlm_input))

        hessian_form = ufl.algorithms.expand_derivatives(d2Fdm2 + dFdm_adj2 + d2Fdudm)
        hessian_output = 0
        if not hessian_form.empty():
            hessian_output -= assemble_adjoint_value(hessian_form)

        if isinstance(c, dolfin.function.expression.BaseExpression):
            return [(hessian_output, W)]
        else:
            return hessian_output

    def prepare_recompute_component(self, inputs, relevant_outputs):
        return self._replace_recompute_form()

    def _replace_recompute_form(self):
        func = self._create_initial_guess()

        bcs = self._recover_bcs()
        lhs = self._replace_form(self.lhs, func=func)
        rhs = 0
        if self.linear:
            rhs = self._replace_form(self.rhs)

        return lhs, rhs, func, bcs

    def _forward_solve(self, lhs, rhs, func, bcs):
        dolfin.solve(lhs == rhs, func, bcs, *self.forward_args, **self.forward_kwargs)
        return func

    def _assembled_solve(self, lhs, rhs, func, bcs, **kwargs):
        for bc in bcs:
            bc.apply(rhs)
        dolfin.solve(lhs, func.vector(), rhs, **kwargs)
        return func

    def recompute_component(self, inputs, block_variable, idx, prepared):
        lhs = prepared[0]
        rhs = prepared[1]
        func = prepared[2]
        bcs = prepared[3]
        return self._forward_solve(lhs, rhs, func, bcs)


class SolveLinearSystemBlock(GenericSolveBlock):
    def __init__(self, A, u, b, *args, **kwargs):
        lhs = A.form
        func = u.function
        rhs = b.form
        bcs = A.bcs if hasattr(A, "bcs") else []
        super().__init__(lhs, rhs, func, bcs, *args, **kwargs)

        # Set up parameters initialization
        self.assemble_kwargs["keep_diagonal"] = A.keep_diagonal if hasattr(A, "keep_diagonal") else False
        self.ident_zeros_tol = A.ident_zeros_tol if hasattr(A, "ident_zeros_tol") else None
        self.assemble_system = A.assemble_system if hasattr(A, "assemble_system") else False

    def _init_solver_parameters(self, args, kwargs):
        super()._init_solver_parameters(args, kwargs)
        if len(self.forward_args) <= 0:
            self.forward_args = args

        if len(self.adj_args) <= 0:
            self.adj_args = self.forward_args

    def _assemble_and_solve_adj_eq(self, dFdu_adj_form, dJdu, compute_bdy):
        dJdu_copy = dJdu.copy()
        bcs = self._homogenize_bcs()
        if self.assemble_system:
            rhs_bcs_form = dolfin.inner(dolfin.Function(self.function_space),
                                        dFdu_adj_form.arguments()[0]) * dolfin.dx
            A, _ = dolfin.assemble_system(dFdu_adj_form, rhs_bcs_form, bcs, **self.assemble_kwargs)
        else:
            kwargs = self.assemble_kwargs.copy()
            kwargs["bcs"] = bcs
            A = assemble_adjoint_value(dFdu_adj_form, **kwargs)
        if self.ident_zeros_tol is not None:
            A.ident_zeros(self.ident_zeros_tol)
        [bc.apply(dJdu) for bc in bcs]

        adj_sol = create_function(self.function_space)
        linalg_solve(A, adj_sol.vector(), dJdu, *self.adj_args, **self.adj_kwargs)

        adj_sol_bdy = None
        if compute_bdy:
            adj_sol_bdy = function_from_vector(self.function_space,
                                               dJdu_copy - assemble_adjoint_value(
                                                   dolfin.action(dFdu_adj_form, adj_sol)))

        return adj_sol, adj_sol_bdy

    def _forward_solve(self, lhs, rhs, func, bcs, **kwargs):
        if self.assemble_system:
            A, b = dolfin.assemble_system(lhs, rhs, bcs)
        else:
            assemble_kwargs = self.assemble_kwargs.copy()
            assemble_kwargs["bcs"] = bcs
            A = assemble_adjoint_value(lhs, **assemble_kwargs)
            b = dolfin.assemble(rhs)
            [bc.apply(b) for bc in bcs]

        if self.ident_zeros_tol is not None:
            A.ident_zeros(self.ident_zeros_tol)

        dolfin.solve(A, func.vector(), b, *self.forward_args, **self.forward_kwargs)
        return func


class SolveVarFormBlock(GenericSolveBlock):
    pop_kwargs_keys = GenericSolveBlock.pop_kwargs_keys

    def __init__(self, equation, func, bcs=[], *args, **kwargs):
        lhs = equation.lhs
        rhs = equation.rhs
        super().__init__(lhs, rhs, func, bcs, *args, **kwargs)

    def _init_solver_parameters(self, args, kwargs):
        super()._init_solver_parameters(args, kwargs)
        if len(self.forward_args) <= 0:
            self.forward_args = args

        if len(self.forward_kwargs) <= 0:
            self.forward_kwargs = kwargs

        if "solver_parameters" in self.forward_kwargs and "mat_type" in self.forward_kwargs["solver_parameters"]:
            self.assemble_kwargs["mat_type"] = self.forward_kwargs["solver_parameters"]["mat_type"]

        if len(self.adj_kwargs) <= 0:
            solver_parameters = kwargs.get("solver_parameters", {})
            if len(self.adj_args) <= 0:
                if "linear_solver" in solver_parameters:
                    adj_args = [solver_parameters["linear_solver"]]
                    if "preconditioner" in solver_parameters:
                        adj_args.append(solver_parameters["preconditioner"])
                    self.adj_args = tuple(adj_args)
                elif "newton_solver" in solver_parameters and "linear_solver" in solver_parameters["newton_solver"]:
                    adj_args = [solver_parameters["newton_solver"]["linear_solver"]]
                    if "preconditioner" in solver_parameters["newton_solver"]:
                        adj_args.append(solver_parameters["newton_solver"]["preconditioner"])
                    self.adj_args = tuple(adj_args)
            self.adj_kwargs = solver_parameters

    def _assemble_and_solve_adj_eq(self, dFdu_adj_form, dJdu, compute_bdy=True):
        dJdu_copy = dJdu.copy()
        bcs = self._homogenize_bcs()
        kwargs = self.assemble_kwargs.copy()
        kwargs["bcs"] = bcs
        dFdu = assemble_adjoint_value(dFdu_adj_form, **kwargs)

        # Apply boundary conditions on adj_dFdu and dJdu.
        for bc in bcs:
            bc.apply(dJdu)

        adj_sol = create_function(self.function_space)
        lu_solver_methods = dolfin.lu_solver_methods()
        solver_method = self.adj_args[0] if len(self.adj_args) >= 1 else "default"
        solver_method = "default" if solver_method == "lu" else solver_method

        if solver_method in lu_solver_methods:
            solver = dolfin.LUSolver(solver_method)
            solver_parameters = self.adj_kwargs.get("lu_solver", {})
        else:
            solver = dolfin.KrylovSolver(*self.adj_args)
            solver_parameters = self.adj_kwargs.get("krylov_solver", {})
        solver.parameters.update(solver_parameters)
        solver.solve(dFdu, adj_sol.vector(), dJdu)

        adj_sol_bdy = None
        if compute_bdy:
            adj_sol_bdy = function_from_vector(self.function_space,
                                               dJdu_copy - assemble_adjoint_value(
                                                   dolfin.action(dFdu_adj_form, adj_sol)))

        return adj_sol, adj_sol_bdy



import dolfin
import ufl_legacy as ufl
import ufl_legacy.algorithms as ufl_algorithms
from pyadjoint.optimization.constraints import (Constraint, EqualityConstraint,
                                                InequalityConstraint)

import fenics_adjoint.types


def as_vec(x):
    if isinstance(x, dolfin.Function):
        out = x.vector().get_local()
    else:
        out = x.get_local()

    if len(out) == 1:
        out = out[0]
    return fenics_adjoint.types.Constant(out)


class UFLConstraint(Constraint):
    """
    Easily implement scalar constraints using UFL.

    The form must be a 0-form that depends on a Function control.
    """

    def __init__(self, form, control):

        if not isinstance(control.control, dolfin.Function):
            raise NotImplementedError("Only implemented for Function controls")

        args = ufl_algorithms.extract_arguments(form)
        if len(args) != 0:
            raise ValueError("Must be a rank-zero form, i.e. a functional")

        u = control.control
        self.V = u.function_space()
        # We want to make a copy of the control purely for use
        # in the constraint, so that our writing it isn't
        # bothering anyone else
        self.u = fenics_adjoint.types.Function(self.V)
        self.form = ufl.replace(form, {u: self.u})

        self.trial = dolfin.TrialFunction(self.V)
        self.dform = dolfin.derivative(self.form, self.u, self.trial)
        if len(ufl_algorithms.extract_arguments(ufl_algorithms.expand_derivatives(self.dform))) == 0:
            raise ValueError("Form must depend on control")

        self.test = dolfin.TestFunction(self.V)
        self.hess = ufl_algorithms.expand_derivatives(dolfin.derivative(self.dform, self.u, self.test))
        if len(ufl_algorithms.extract_arguments(self.hess)) == 0:
            self.zero_hess = True
        else:
            self.zero_hess = False

    def update_control(self, m):
        if isinstance(m, list):
            assert len(m) == 1
            m = m[0]

        if isinstance(m, dolfin.Function):
            self.u.assign(m)
        else:
            self.u._ad_assign_numpy(self.u, m, 0)

    def function(self, m):
        self.update_control(m)
        b = dolfin.assemble(self.form)
        return fenics_adjoint.types.Constant(b)

    def jacobian(self, m):
        if isinstance(m, list):
            assert len(m) == 1
            m = m[0]

        self.update_control(m)
        out = [dolfin.assemble(self.dform)]
        return out

    def jacobian_action(self, m, dm, result):
        """Computes the Jacobian action of c(m) in direction dm and stores the result in result. """

        if isinstance(m, list):
            assert len(m) == 1
            m = m[0]
        self.update_control(m)

        form = dolfin.action(self.dform, dm)
        result.assign(dolfin.assemble(form))

    def jacobian_adjoint_action(self, m, dp, result):
        """Computes the Jacobian adjoint action of c(m) in direction dp and stores the result in result. """

        if isinstance(m, list):
            assert len(m) == 1
            m = m[0]
        self.update_control(m)

        asm = dolfin.assemble(dp * ufl.replace(self.dform, {self.trial: self.test}))
        if isinstance(result, dolfin.Function):
            if dolfin.__name__ in ["dolfin", "fenics"]:
                result.vector().zero()
                result.vector().axpy(1.0, asm)
            else:
                result.assign(asm)
        else:
            raise NotImplementedError("Do I need to untangle all controls?")

    def hessian_action(self, m, dm, dp, result):
        """Computes the Hessian action of c(m) in direction dm and dp and stores the result in result. """

        if isinstance(m, list):
            assert len(m) == 1
            m = m[0]
        self.update_control(m)

        H = dm * ufl.replace(self.hess, {self.trial: dp})
        if isinstance(result, dolfin.Function):
            if dolfin.__name__ in ["dolfin", "fenics"]:
                if self.zero_hess:
                    result.vector().zero()
                else:
                    result.vector().zero()
                    result.vector().axpy(1.0, dolfin.assemble(H))
            else:
                if self.zero_hess:
                    result.assign(0)
                else:
                    result.assign(dolfin.assemble(H))

        else:
            raise NotImplementedError("Do I need to untangle all controls?")

    def output_workspace(self):
        """Return an object like the output of c(m) for calculations."""

        return fenics_adjoint.types.Constant(dolfin.assemble(self.form))

    def _get_constraint_dim(self):
        """Returns the number of constraint components."""
        return 1


class UFLEqualityConstraint(UFLConstraint, EqualityConstraint):
    pass


class UFLInequalityConstraint(UFLConstraint, InequalityConstraint):
    pass



# flake8: noqa

from .solving import SolveLinearSystemBlock, SolveVarFormBlock
from .projection import ProjectBlock
from .variational_solver import LinearVariationalSolveBlock, NonlinearVariationalSolveBlock
from .krylov_solver import KrylovSolveBlock, KrylovSolveBlockHelper
from .lu_solver import LUSolveBlock, LUSolveBlockHelper
from .function_assigner import FunctionAssignerBlock
from .function import FunctionEvalBlock, FunctionSplitBlock, FunctionMergeBlock
from .petsc_krylov_solver import PETScKrylovSolveBlock, PETScKrylovSolveBlockHelper
from .constant import ConstantAssignBlock
from .dirichlet_bc import DirichletBCBlock
from .assembly import AssembleBlock



"""This just acts as an alias for fenics_adjoint.

"""
# flake8: noqa

from fenics_adjoint import *

from importlib.metadata import metadata

meta = metadata("dolfin_adjoint")

__version__ = meta["Version"]
__author__ = meta["Author"]
__license__ = meta["License"]



import dolfin
from pyadjoint.tape import get_working_tape, stop_annotating, annotate_tape
from pyadjoint.overloaded_type import create_overloaded_object
from .blocks import AssembleBlock


def assemble(*args, **kwargs):
    """When a form is assembled, the information about its nonlinear dependencies is lost,
    and it is no longer easy to manipulate. Therefore, fenics_adjoint overloads the :py:func:`dolfin.assemble`
    function to *attach the form to the assembled object*. This lets the automatic annotation work,
    even when the user calls the lower-level :py:data:`solve(A, x, b)`.
    """
    ad_block_tag = kwargs.pop("ad_block_tag", None)
    annotate = annotate_tape(kwargs)
    with stop_annotating():
        output = dolfin.assemble(*args, **kwargs)
    if "keep_diagonal" in kwargs:
        output.keep_diagonal = kwargs["keep_diagonal"]

    form = args[0]
    if isinstance(output, float):
        output = create_overloaded_object(output)

        if annotate:
            block = AssembleBlock(form, ad_block_tag=ad_block_tag)

            tape = get_working_tape()
            tape.add_block(block)

            block.add_output(output.block_variable)
    else:
        # Assembled a vector or matrix
        output.form = form

    return output


def assemble_system(*args, **kwargs):
    """When a form is assembled, the information about its nonlinear dependencies is lost,
    and it is no longer easy to manipulate. Therefore, fenics_adjoint overloads the :py:func:`dolfin.assemble_system`
    function to *attach the form to the assembled object*. This lets the automatic annotation work,
    even when the user calls the lower-level :py:data:`solve(A, x, b)`.
    """
    A_form = args[0]
    b_form = args[1]

    A, b = dolfin.assemble_system(*args, **kwargs)
    if "keep_diagonal" in kwargs:
        A.keep_diagonal = kwargs["keep_diagonal"]
    if "bcs" in kwargs:
        bcs = kwargs["bcs"]
    elif len(args) > 2:
        bcs = args[2]
    else:
        bcs = []

    A.form = A_form
    A.bcs = bcs
    b.form = b_form
    b.bcs = bcs
    A.assemble_system = True

    return A, b



import dolfin
from .solving import SolveVarFormBlock


class LinearVariationalSolveBlock(SolveVarFormBlock):
    def __init__(self, equation, func, bcs,
                 problem_args, problem_kwargs,
                 solver_params, solver_args,
                 solver_kwargs, solve_args, solve_kwargs,
                 **kwargs):
        func = func
        bcs = bcs

        self.problem_args = problem_args
        self.problem_kwargs = problem_kwargs
        self.solver_params = solver_params.copy()
        self.solver_args = solver_args
        self.solver_kwargs = solver_kwargs
        self.solve_args = solve_args
        self.solve_kwargs = solve_kwargs

        super(LinearVariationalSolveBlock, self).__init__(equation, func, bcs, **kwargs)

    def _init_solver_parameters(self, args, kwargs):
        super()._init_solver_parameters(args, kwargs)
        if len(self.adj_args) <= 0 and len(self.adj_kwargs) <= 0:
            method = self.solver_params["linear_solver"]
            precond = self.solver_params["preconditioner"]
            self.adj_args = [method, precond]

    def _forward_solve(self, lhs, rhs, func, bcs, **kwargs):
        problem = dolfin.LinearVariationalProblem(lhs, rhs, func, bcs, *self.problem_args, **self.problem_kwargs)
        solver = dolfin.LinearVariationalSolver(problem, *self.solver_args, **self.solver_kwargs)
        solver.parameters.update(self.solver_params)
        solver.solve(*self.solve_args, **self.solve_kwargs)
        return func


class NonlinearVariationalSolveBlock(SolveVarFormBlock):
    def __init__(self, equation, func, bcs, problem_J,
                 problem_args, problem_kwargs,
                 solver_params, solver_args,
                 solver_kwargs, solve_args, solve_kwargs,
                 **kwargs):
        func = func
        bcs = bcs

        self.problem_J = problem_J
        self.problem_args = problem_args
        self.problem_kwargs = problem_kwargs
        self.solver_params = solver_params.copy()
        self.solver_args = solver_args
        self.solver_kwargs = solver_kwargs
        self.solve_args = solve_args
        self.solve_kwargs = solve_kwargs

        super(NonlinearVariationalSolveBlock, self).__init__(equation, func, bcs, **kwargs)

        if self.problem_J is not None:
            for coeff in self.problem_J.coefficients():
                self.add_dependency(coeff, no_duplicates=True)

    def _init_solver_parameters(self, args, kwargs):
        super()._init_solver_parameters(args, kwargs)
        if len(self.adj_args) <= 0 and len(self.adj_kwargs) <= 0:
            params_key = "{}_solver".format(self.solver_params["nonlinear_solver"])

            if params_key in self.solver_params:
                method = self.solver_params[params_key]["linear_solver"]
                precond = self.solver_params[params_key]["preconditioner"]
                self.adj_args = [method, precond]

    def _forward_solve(self, lhs, rhs, func, bcs, **kwargs):
        J = self.problem_J
        if J is not None:
            J = self._replace_form(J, func)
        problem = dolfin.NonlinearVariationalProblem(lhs, func, bcs, J=J,
                                                     *self.problem_args, **self.problem_kwargs)
        solver = dolfin.NonlinearVariationalSolver(problem, *self.solver_args, **self.solver_kwargs)
        solver.parameters.update(self.solver_params)
        solver.solve(*self.solve_args, **self.solve_kwargs)
        return func



"""

The entire dolfin-adjoint interface should be imported with a single
call:

.. code-block:: python

  from dolfin import *
  from dolfin_adjoint import *

It is essential that the importing of the :py:mod:`dolfin_adjoint` module happen *after*
importing the :py:mod:`dolfin` module. dolfin-adjoint relies on *overloading* many of
the key functions of dolfin to achieve its degree of automation.
"""

# flake8: noqa

from pyadjoint import (Tape, set_working_tape, get_working_tape,
                       pause_annotation, continue_annotation,
                       ReducedFunctional,
                       taylor_test, taylor_to_dict,
                       compute_gradient, compute_hessian,
                       AdjFloat, Control, minimize, maximize, MinimizationProblem,
                       IPOPTSolver, ROLSolver, InequalityConstraint, EqualityConstraint,
                       MoolaOptimizationProblem, print_optimization_methods,
                       stop_annotating)
from .variational_solver import (NonlinearVariationalProblem, NonlinearVariationalSolver,
                                 LinearVariationalProblem, LinearVariationalSolver)
from .system_assembly import *
from .refine import refine
from .types import *
from .petsc_krylov_solver import PETScKrylovSolver
from .krylov_solver import KrylovSolver
from .lu_solver import LUSolver
from .newton_solver import NewtonSolver
from .ufl_constraints import UFLInequalityConstraint, UFLEqualityConstraint
from .shapead_transformations import (transfer_from_boundary,
                                      transfer_to_boundary)
from .interpolation import interpolate
from .projection import project
from .solving import solve
from .assembly import assemble, assemble_system

import pyadjoint

import sys
import dolfin


from importlib.metadata import metadata

meta = metadata("dolfin_adjoint")
__version__ = meta["Version"]
__author__ = meta["Author"]
__license__ = meta["License"]

set_working_tape(Tape())
continue_annotation()



import dolfin
from pyadjoint.overloaded_type import create_overloaded_object
from pyadjoint.tape import stop_annotating


def refine(*args, **kwargs):
    """ Refine is overloaded to ensure that the returned mesh is overloaded.
    """
    kwargs.pop("ad_block_tag", None)
    with stop_annotating():
        output = dolfin.refine(*args, **kwargs)
    overloaded = create_overloaded_object(output)
    return overloaded



import dolfin

from fenics_adjoint.utils import function_from_vector


from . import SolveLinearSystemBlock
from .assembly import assemble_adjoint_value


class KrylovSolveBlockHelper(object):
    def __init__(self):
        self.forward_solver = None
        self.adjoint_solver = None

    def reset(self):
        self.forward_solver = None
        self.adjoint_solver = None


class KrylovSolveBlock(SolveLinearSystemBlock):
    def __init__(self, A, u, b,
                 krylov_solver_parameters,
                 block_helper, nonzero_initial_guess,
                 pc_operator, krylov_method,
                 krylov_preconditioner,
                 **kwargs):
        super(KrylovSolveBlock, self).__init__(A, u, b, **kwargs)

        self.krylov_solver_parameters = krylov_solver_parameters
        self.block_helper = block_helper
        self.nonzero_initial_guess = nonzero_initial_guess
        self.pc_operator = pc_operator
        self.method = krylov_method
        self.preconditioner = krylov_preconditioner

        if self.nonzero_initial_guess:
            # Here we store a variable that isn't necessarily a dependency.
            # This means that the graph does not know that we depend on this BlockVariable.
            # This could lead to unexpected behaviour in the future.
            # TODO: Consider if this is really a problem.
            self.func.block_variable.save_output()
            self.initial_guess = self.func.block_variable

        if self.pc_operator is not None:
            self.pc_operator = self.pc_operator.form
            for c in self.pc_operator.coefficients():
                self.add_dependency(c)

    def _create_initial_guess(self):
        r = super(KrylovSolveBlock, self)._create_initial_guess()
        if self.nonzero_initial_guess:
            dolfin.Function.assign(r, self.initial_guess.saved_output)
        return r

    def _assemble_and_solve_adj_eq(self, dFdu_adj_form, dJdu, compute_bdy):
        dJdu_copy = dJdu.copy()
        bcs = self._homogenize_bcs()

        solver = self.block_helper.adjoint_solver
        if solver is None:
            solver = dolfin.KrylovSolver(self.method, self.preconditioner)

            if self.assemble_system:
                rhs_bcs_form = dolfin.inner(dolfin.Function(self.function_space),
                                            dFdu_adj_form.arguments()[0]) * dolfin.dx
                A, _ = dolfin.assemble_system(dFdu_adj_form, rhs_bcs_form, bcs)

                if self.pc_operator is not None:
                    P = self._replace_form(self.pc_operator)
                    P, _ = dolfin.assemble_system(P, rhs_bcs_form, bcs)
                    solver.set_operators(A, P)
                else:
                    solver.set_operator(A)
            else:
                A = assemble_adjoint_value(dFdu_adj_form)
                [bc.apply(A) for bc in bcs]

                if self.pc_operator is not None:
                    P = self._replace_form(self.pc_operator)
                    P = assemble_adjoint_value(P)
                    [bc.apply(P) for bc in bcs]
                    solver.set_operators(A, P)
                else:
                    solver.set_operator(A)

            self.block_helper.adjoint_solver = solver

        solver.parameters.update(self.krylov_solver_parameters)
        [bc.apply(dJdu) for bc in bcs]

        adj_sol = dolfin.Function(self.function_space)
        solver.solve(adj_sol.vector(), dJdu)

        adj_sol_bdy = None
        if compute_bdy:
            adj_sol_bdy = function_from_vector(self.function_space, dJdu_copy - assemble_adjoint_value(
                dolfin.action(dFdu_adj_form, adj_sol)))

        return adj_sol, adj_sol_bdy

    def _forward_solve(self, lhs, rhs, func, bcs, **kwargs):
        solver = self.block_helper.forward_solver
        if solver is None:
            solver = dolfin.KrylovSolver(self.method, self.preconditioner)
            if self.assemble_system:
                A, _ = dolfin.assemble_system(lhs, rhs, bcs)
                if self.pc_operator is not None:
                    P = self._replace_form(self.pc_operator)
                    P, _ = dolfin.assemble_system(P, rhs, bcs)
                    solver.set_operators(A, P)
                else:
                    solver.set_operator(A)
            else:
                A = assemble_adjoint_value(lhs)
                [bc.apply(A) for bc in bcs]
                if self.pc_operator is not None:
                    P = self._replace_form(self.pc_operator)
                    P = assemble_adjoint_value(P)
                    [bc.apply(P) for bc in bcs]
                    solver.set_operators(A, P)
                else:
                    solver.set_operator(A)
            self.block_helper.forward_solver = solver

        if self.assemble_system:
            system_assembler = dolfin.SystemAssembler(lhs, rhs, bcs)
            b = dolfin.Function(self.function_space).vector()
            system_assembler.assemble(b)
        else:
            b = assemble_adjoint_value(rhs)
            [bc.apply(b) for bc in bcs]

        solver.parameters.update(self.krylov_solver_parameters)
        solver.solve(func.vector(), b)
        return func



import dolfin
from pyadjoint.tape import get_working_tape, annotate_tape, stop_annotating
from pyadjoint.overloaded_type import create_overloaded_object
from .blocks import ProjectBlock


def project(*args, **kwargs):
    """The project call performs an equation solve, and so it too must be annotated so that the
    adjoint and tangent linear models may be constructed automatically by pyadjoint.

    To disable the annotation of this function, just pass :py:data:`annotate=False`. This is useful in
    cases where the solve is known to be irrelevant or diagnostic for the purposes of the adjoint
    computation (such as projecting fields to other function spaces for the purposes of
    visualisation)."""

    ad_block_tag = kwargs.pop("ad_block_tag", None)
    annotate = annotate_tape(kwargs)
    with stop_annotating():
        output = dolfin.project(*args, **kwargs)
    output = create_overloaded_object(output)

    if annotate:
        bcs = kwargs.pop("bcs", [])
        sb_kwargs = ProjectBlock.pop_kwargs(kwargs)
        sb_kwargs.update(kwargs)
        block = ProjectBlock(args[0], args[1], output, bcs,
                             ad_block_tag=ad_block_tag, **sb_kwargs)

        tape = get_working_tape()
        tape.add_block(block)

        block.add_output(output.block_variable)

    return output



import dolfin
from pyadjoint.tape import get_working_tape, stop_annotating, annotate_tape, no_annotations
from .blocks import LinearVariationalSolveBlock, NonlinearVariationalSolveBlock


class NonlinearVariationalProblem(dolfin.NonlinearVariationalProblem):
    """This object is overloaded so that solves using this class are automatically annotated,
    so that pyadjoint can automatically derive the adjoint and tangent linear models."""

    @no_annotations
    def __init__(self, F, u, bcs=None, J=None, *args, **kwargs):
        super(NonlinearVariationalProblem, self).__init__(F, u, bcs, J,
                                                          *args, **kwargs)
        self._ad_F = F
        self._ad_u = u
        self._ad_bcs = bcs
        self._ad_J = J
        self._ad_args = args
        self._ad_kwargs = kwargs


class NonlinearVariationalSolver(dolfin.NonlinearVariationalSolver):
    """This object is overloaded so that solves using this class are automatically annotated,
    so that pyadjoint can automatically derive the adjoint and tangent linear models."""

    @no_annotations
    def __init__(self, problem, *args, **kwargs):
        super(NonlinearVariationalSolver, self).__init__(problem, *args, **kwargs)
        self.ad_block_tag = kwargs.pop("ad_block_tag", None)
        self._ad_problem = problem
        self._ad_args = args
        self._ad_kwargs = kwargs

    def solve(self, *args, **kwargs):
        """To disable the annotation, just pass :py:data:`annotate=False` to this routine, and it acts exactly like the
        Dolfin solve call. This is useful in cases where the solve is known to be irrelevant or diagnostic
        for the purposes of the adjoint computation (such as projecting fields to other function spaces
        for the purposes of visualisation)."""

        annotate = annotate_tape(kwargs)
        if annotate:
            tape = get_working_tape()
            problem = self._ad_problem
            sb_kwargs = NonlinearVariationalSolveBlock.pop_kwargs(kwargs)
            sb_kwargs.update(kwargs)
            block = NonlinearVariationalSolveBlock(problem._ad_F == 0,
                                                   problem._ad_u,
                                                   problem._ad_bcs,
                                                   problem_J=problem._ad_J,
                                                   problem_args=problem._ad_args,
                                                   problem_kwargs=problem._ad_kwargs,
                                                   solver_params=self.parameters,
                                                   solver_args=self._ad_args,
                                                   solver_kwargs=self._ad_kwargs,
                                                   solve_args=args,
                                                   solve_kwargs=kwargs,
                                                   ad_block_tag=self.ad_block_tag,
                                                   **sb_kwargs)
            tape.add_block(block)

        with stop_annotating():
            out = super(NonlinearVariationalSolver, self).solve(*args, **kwargs)

        if annotate:
            block.add_output(self._ad_problem._ad_u.create_block_variable())

        return out


class LinearVariationalProblem(dolfin.LinearVariationalProblem):
    """This object is overloaded so that solves using this class are automatically annotated,
    so that pyadjoint can automatically derive the adjoint and tangent linear models."""

    @no_annotations
    def __init__(self, a, L, u, bcs=None, *args, **kwargs):
        super(LinearVariationalProblem, self).__init__(a, L, u, bcs,
                                                       *args, **kwargs)
        self._ad_a = a
        self._ad_L = L
        self._ad_u = u
        self._ad_bcs = bcs
        self._ad_args = args
        self._ad_kwargs = kwargs


class LinearVariationalSolver(dolfin.LinearVariationalSolver):
    """This object is overloaded so that solves using this class are automatically annotated,
    so that pyadjoint can automatically derive the adjoint and tangent linear models."""

    @no_annotations
    def __init__(self, problem, *args, **kwargs):
        super(LinearVariationalSolver, self).__init__(problem, *args, **kwargs)
        self.ad_block_tag = kwargs.pop("ad_block_tag", None)
        self._ad_problem = problem
        self._ad_args = args
        self._ad_kwargs = kwargs

    def solve(self, *args, **kwargs):
        """To disable the annotation, just pass :py:data:`annotate=False` to this routine, and it acts exactly like the
        Dolfin solve call. This is useful in cases where the solve is known to be irrelevant or diagnostic
        for the purposes of the adjoint computation (such as projecting fields to other function spaces
        for the purposes of visualisation)."""

        annotate = annotate_tape(kwargs)
        if annotate:
            tape = get_working_tape()
            problem = self._ad_problem
            sb_kwargs = LinearVariationalSolveBlock.pop_kwargs(kwargs)
            sb_kwargs.update(kwargs)
            block = LinearVariationalSolveBlock(problem._ad_a == problem._ad_L,
                                                problem._ad_u,
                                                problem._ad_bcs,
                                                problem_args=problem._ad_args,
                                                problem_kwargs=problem._ad_kwargs,
                                                solver_params=self.parameters,
                                                solver_args=self._ad_args,
                                                solver_kwargs=self._ad_kwargs,
                                                solve_args=args,
                                                solve_kwargs=kwargs,
                                                ad_block_tag=self.ad_block_tag,
                                                **sb_kwargs)
            tape.add_block(block)

        with stop_annotating():
            out = super(LinearVariationalSolver, self).solve(*args, **kwargs)

        if annotate:
            block.add_output(self._ad_problem._ad_u.create_block_variable())

        return out



import dolfin
from pyadjoint.tape import annotate_tape, get_working_tape


from .blocks import LUSolveBlock, LUSolveBlockHelper
from fenics_adjoint.utils import MatrixTypes


class LUSolver(dolfin.LUSolver):
    def __init__(self, *args, **kwargs):
        self.ad_block_tag = kwargs.pop("ad_block_tag", None)
        dolfin.LUSolver.__init__(self, *args, **kwargs)

        A = kwargs.pop("A", None)
        method = kwargs.pop("method", "default")

        next_arg_idx = 0
        if len(args) > 0 and isinstance(args[0], MatrixTypes):
            A = args[0]
            next_arg_idx = 1
        elif len(args) > 1 and isinstance(args[1], MatrixTypes):
            A = args[1]
            next_arg_idx = 2

        if len(args) > next_arg_idx and isinstance(args[next_arg_idx], str):
            method = args[next_arg_idx]

        self.operator = A
        self.method = method
        self.solver_parameters = {}
        self.block_helper = LUSolveBlockHelper()

    def set_operator(self, arg0):
        self.operator = arg0
        self.block_helper = LUSolveBlockHelper()
        return dolfin.LUSolver.set_operator(self, arg0)

    def solve(self, *args, **kwargs):
        annotate = annotate_tape(kwargs)

        if annotate:
            if len(args) == 3:
                block_helper = LUSolveBlockHelper()
                A = args[0]
                x = args[1]
                b = args[2]
            elif len(args) == 2:
                block_helper = self.block_helper
                A = self.operator
                x = args[0]
                b = args[1]

            u = x.function
            parameters = self.parameters.copy()

            tape = get_working_tape()
            sb_kwargs = LUSolveBlock.pop_kwargs(kwargs)
            block = LUSolveBlock(A, x, b,
                                 lu_solver_parameters=parameters,
                                 block_helper=block_helper,
                                 lu_solver_method=self.method,
                                 ad_block_tag=self.ad_block_tag,
                                 **sb_kwargs)
            tape.add_block(block)

        out = dolfin.LUSolver.solve(self, *args, **kwargs)

        if annotate:
            block.add_output(u.create_block_variable())

        return out



import dolfin
import numpy
from pyadjoint.overloaded_type import (OverloadedType,
                                       create_overloaded_object,
                                       register_overloaded_type)
from pyadjoint.reduced_functional_numpy import gather
from pyadjoint.tape import annotate_tape, get_working_tape, stop_annotating


from fenics_adjoint.blocks import ConstantAssignBlock
from fenics_adjoint.utils import constant_from_values


def create_constant(*args, **kwargs):
    """Initialise a fenics_adjoint.Constant object and return it."""
    # Dolfin constants do not have domains
    _ = kwargs.pop("domain", None)
    return Constant(*args, **kwargs)


@register_overloaded_type
class Constant(OverloadedType, dolfin.Constant):
    def __init__(self, *args, **kwargs):
        ad_block_tag = kwargs.pop("ad_block_tag", None)
        annotate = annotate_tape(kwargs)
        super(Constant, self).__init__(*args, **kwargs)
        dolfin.Constant.__init__(self, *args, **kwargs)

        if annotate and len(args) > 0:
            value = args[0]
            if isinstance(value, OverloadedType):
                block = ConstantAssignBlock(value, ad_block_tag=ad_block_tag)
                tape = get_working_tape()
                tape.add_block(block)
                block.add_output(self.block_variable)
            elif isinstance(value, (tuple, list)):
                value = numpy.array(value, dtype="O")
                if any(isinstance(v, OverloadedType) for v in value.flat):
                    block = ConstantAssignBlock(value, ad_block_tag=ad_block_tag)
                    tape = get_working_tape()
                    tape.add_block(block)
                    block.add_output(self.block_variable)

    def assign(self, *args, **kwargs):
        ad_block_tag = kwargs.pop("ad_block_tag", None)
        annotate = annotate_tape(kwargs)
        if annotate:
            other = args[0]
            if not isinstance(other, OverloadedType):
                other = create_overloaded_object(other)

            block = ConstantAssignBlock(other, ad_block_tag=ad_block_tag)
            tape = get_working_tape()
            tape.add_block(block)

        with stop_annotating():
            ret = dolfin.Constant.assign(self, *args, **kwargs)

        if annotate:
            block.add_output(self.create_block_variable())

        return ret

    def get_derivative(self, options={}):
        return self._ad_convert_type(self.adj_value, options=options)

    @classmethod
    def _ad_init_object(cls, obj):
        # In FEniCS, passing a Constant to the Constant constructor is not possible when the Constant is nonscalar.
        values = obj.values()
        shape = obj.ufl_shape
        return cls(numpy.reshape(values, shape))

    def _ad_convert_type(self, value, options={}):
        if value is None:
            # TODO: Should the default be 0 constant here or return just None?
            return Constant(numpy.zeros(self.ufl_shape))
        value = gather(value)
        return constant_from_values(self, value)

    def _ad_function_space(self, mesh):
        element = self.ufl_element()
        fs_element = element.reconstruct(cell=mesh.ufl_cell())
        return dolfin.FunctionSpace(mesh, fs_element)

    def _ad_create_checkpoint(self):
        return constant_from_values(self)

    def _ad_restore_at_checkpoint(self, checkpoint):
        return checkpoint

    def _ad_mul(self, other):
        return constant_from_values(self, self.values() * other)

    def _ad_add(self, other):
        return constant_from_values(self, self.values() + other.values())

    def _ad_dot(self, other, options=None):
        return sum(self.values() * other.values())

    @staticmethod
    def _ad_assign_numpy(dst, src, offset):
        dst.assign(dolfin.Constant(numpy.reshape(src[offset:offset + dst.value_size()], dst.ufl_shape)))
        offset += dst.value_size()
        return dst, offset

    @staticmethod
    def _ad_to_list(m):
        a = numpy.zeros(m.value_size())
        p = numpy.zeros(m.value_size())
        m.eval(a, p)
        return a.tolist()

    def _ad_copy(self):
        return constant_from_values(self)

    def _ad_dim(self):
        return numpy.prod(self.values().shape)

    def _ad_imul(self, other):
        self.assign(constant_from_values(self, self.values() * other))

    def _ad_iadd(self, other):
        self.assign(constant_from_values(self, self.values() + other.values()))

    def _reduce(self, r, r0):
        npdata = self.values()
        for i in range(len(npdata)):
            r0 = r(npdata[i], r0)
        return r0

    def _applyUnary(self, f):
        npdata = self.values()
        npdatacopy = npdata.copy()
        for i in range(len(npdata)):
            npdatacopy[i] = f(npdata[i])
        self.assign(constant_from_values(self, npdatacopy))

    def _applyBinary(self, f, y):
        npdata = self.values()
        npdatacopy = self.values().copy()
        npdatay = y.values()
        for i in range(len(npdata)):
            npdatacopy[i] = f(npdata[i], npdatay[i])
        self.assign(constant_from_values(self, npdatacopy))

    def __deepcopy__(self, memodict={}):
        return constant_from_values(self)



import dolfin
import ufl_legacy as ufl
from pyadjoint import Block, create_overloaded_object
from ufl_legacy.formatting.ufl2unicode import ufl2unicode
from fenics_adjoint.utils import extract_mesh_from_form


def assemble_adjoint_value(*args, **kwargs):
    """Wrapper that assembles a matrix with boundary conditions"""
    bcs = kwargs.pop("bcs", ())
    result = dolfin.assemble(*args, **kwargs)
    for bc in bcs:
        bc.apply(result)
    return result


class AssembleBlock(Block):
    def __init__(self, form, ad_block_tag=None):
        super(AssembleBlock, self).__init__(ad_block_tag=ad_block_tag)
        self.form = form
        mesh = self.form.ufl_domain().ufl_cargo()
        self.add_dependency(mesh)
        for c in self.form.coefficients():
            self.add_dependency(c, no_duplicates=True)

    def __str__(self):
        return f"assemble({ufl2unicode(self.form)})"

    def compute_action_adjoint(self, adj_input, arity_form, form=None, c_rep=None, space=None, dform=None):
        """This computes the action of the adjoint of the derivative of `form` wrt `c_rep` on `adj_input`.
           In other words, it returns: `<(dform/dc_rep)*, adj_input>`

           - If `form` has arity 0 => `dform/dc_rep` is a 1-form and `adj_input` a foat, we can simply use
             the `*` operator.

           - If `form` has arity 1 => `dform/dc_rep` is a 2-form and we can symbolically take its adjoint
             and then apply the action on `adj_input`, to finally assemble the result.
        """
        if arity_form == 0:
            if dform is None:
                dc = dolfin.TestFunction(space)
                dform = dolfin.derivative(form, c_rep, dc)
            dform_vector = assemble_adjoint_value(dform)
            # Return a Vector scaled by the scalar `adj_input`
            return dform_vector * adj_input, dform
        elif arity_form == 1:
            if dform is None:
                dc = dolfin.TrialFunction(space)
                dform = dolfin.derivative(form, c_rep, dc)
            # Get the Function
            adj_input = adj_input.function
            # Symbolic operators such as action/adjoint require derivatives to have been expanded beforehand.
            # However, UFL doesn't support expanding coordinate derivatives of Coefficients in physical space,
            # implying that we can't symbolically take the action/adjoint of the Jacobian for SpatialCoordinates.
            # -> Workaround: Apply action/adjoint numerically (using PETSc).
            if not isinstance(c_rep, dolfin.SpatialCoordinate):
                # Symbolically compute: (dform/dc_rep)^* * adj_input
                adj_output = dolfin.action(dolfin.adjoint(dform), adj_input)
                adj_output = assemble_adjoint_value(adj_output)
            else:
                # Get PETSc matrix
                dform_mat = assemble_adjoint_value(dform).petscmat
                # Action of the adjoint (Hermitian transpose)
                adj_output = dolfin.Function(space)
                with adj_input.dat.vec_ro as v_vec:
                    with adj_output.dat.vec as res_vec:
                        dform_mat.multHermitian(v_vec, res_vec)
            return adj_output, dform
        else:
            raise ValueError('Forms with arity > 1 are not handled yet!')

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        replaced_coeffs = {}
        for block_variable in self.get_dependencies():
            coeff = block_variable.output
            c_rep = block_variable.saved_output
            if coeff in self.form.coefficients():
                replaced_coeffs[coeff] = c_rep

        form = ufl.replace(self.form, replaced_coeffs)
        return form

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        form = prepared
        adj_input = adj_inputs[0]
        c = block_variable.output
        c_rep = block_variable.saved_output

        from ufl_legacy.algorithms.analysis import extract_arguments
        arity_form = len(extract_arguments(form))

        if isinstance(c, dolfin.function.expression.BaseExpression):
            # Create a FunctionSpace from self.form and Expression.
            # And then make a TestFunction from this space.
            mesh = self.form.ufl_domain().ufl_cargo()
            V = c._ad_function_space(mesh)
            dc = dolfin.TestFunction(V)

            dform = dolfin.derivative(form, c_rep, dc)
            output = assemble_adjoint_value(dform)
            return [[adj_input * output, V]]

        if isinstance(c, dolfin.Constant):
            mesh = extract_mesh_from_form(self.form)
            space = c._ad_function_space(mesh)
        elif isinstance(c, dolfin.Function):
            space = c.function_space()
        elif isinstance(c, dolfin.Mesh):
            c_rep = dolfin.SpatialCoordinate(c_rep)
            space = c._ad_function_space()

        return self.compute_action_adjoint(adj_input, arity_form, form, c_rep, space)[0]

    def prepare_evaluate_tlm(self, inputs, tlm_inputs, relevant_outputs):
        return self.prepare_evaluate_adj(inputs, tlm_inputs, self.get_dependencies())

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        form = prepared
        dform = 0.

        from ufl_legacy.algorithms.analysis import extract_arguments
        arity_form = len(extract_arguments(form))
        for bv in self.get_dependencies():
            c_rep = bv.saved_output
            tlm_value = bv.tlm_value

            if tlm_value is None:
                continue
            if isinstance(c_rep, dolfin.Mesh):
                X = dolfin.SpatialCoordinate(c_rep)
                dform += dolfin.derivative(form, X, tlm_value)
            else:
                dform += dolfin.derivative(form, c_rep, tlm_value)
        if not isinstance(dform, float):
            dform = ufl.algorithms.expand_derivatives(dform)
            dform = assemble_adjoint_value(dform)
            if arity_form == 1 and dform != 0:
                # Then dform is a Vector
                dform = dform.function
        return dform

    def prepare_evaluate_hessian(self, inputs, hessian_inputs, adj_inputs, relevant_dependencies):
        return self.prepare_evaluate_adj(inputs, adj_inputs, relevant_dependencies)

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        form = prepared
        hessian_input = hessian_inputs[0]
        adj_input = adj_inputs[0]

        from ufl_legacy.algorithms.analysis import extract_arguments
        arity_form = len(extract_arguments(form))

        c1 = block_variable.output
        c1_rep = block_variable.saved_output

        if isinstance(c1, dolfin.Constant):
            mesh = extract_mesh_from_form(form)
            space = c1._ad_function_space(mesh)
        elif isinstance(c1, dolfin.Function):
            space = c1.function_space()
        elif isinstance(c1, dolfin.function.expression.BaseExpression):
            mesh = form.ufl_domain().ufl_cargo()
            space = c1._ad_function_space(mesh)
        elif isinstance(c1, dolfin.Mesh):
            c1_rep = dolfin.SpatialCoordinate(c1)
            space = c1._ad_function_space()
        else:
            return None

        hessian_outputs, dform = self.compute_action_adjoint(hessian_input, arity_form, form, c1_rep, space)

        ddform = 0
        for other_idx, bv in relevant_dependencies:
            c2_rep = bv.saved_output
            tlm_input = bv.tlm_value

            if tlm_input is None:
                continue

            if isinstance(c2_rep, dolfin.Mesh):
                X = dolfin.SpatialCoordinate(c2_rep)
                ddform += dolfin.derivative(dform, X, tlm_input)
            else:
                ddform += dolfin.derivative(dform, c2_rep, tlm_input)

        if not isinstance(ddform, float):
            ddform = ufl.algorithms.expand_derivatives(ddform)
            if not ddform.empty():
                hessian_outputs += self.compute_action_adjoint(adj_input, arity_form, dform=ddform)[0]

        if isinstance(c1, dolfin.function.expression.BaseExpression):
            return [(hessian_outputs, space)]
        else:
            return hessian_outputs

    def prepare_recompute_component(self, inputs, relevant_outputs):
        return self.prepare_evaluate_adj(inputs, None, None)

    def recompute_component(self, inputs, block_variable, idx, prepared):
        form = prepared
        output = dolfin.assemble(form)
        output = create_overloaded_object(output)
        return output



import dolfin
import numpy
from pyadjoint.overloaded_type import create_overloaded_object
from pyadjoint.block import Block
from pyadjoint.tape import get_working_tape, stop_annotating, annotate_tape, no_annotations


def SyncSum(vec):
    """ Returns sum of vec over all mpi processes.

    Each vec vector must have the same dimension for each MPI process """

    comm = dolfin.MPI.comm_world
    NormalsAllProcs = numpy.zeros(comm.Get_size() * len(vec), dtype=vec.dtype)
    comm.Allgather(vec, NormalsAllProcs)

    out = numpy.zeros(len(vec))
    for j in range(comm.Get_size()):
        out += NormalsAllProcs[len(vec) * j:len(vec) * (j + 1)]
    return out


def mesh_to_boundary(v, b_mesh):
    """
    Returns a the boundary representation of the CG-1 function v
    """
    # Extract the underlying volume and boundary meshes
    mesh = v.function_space().mesh()

    # We use a Dof->Vertex mapping to create a global
    # array with all DOF values ordered by mesh vertices
    DofToVert = dolfin.dof_to_vertex_map(v.function_space())
    VGlobal = numpy.zeros(v.vector().size())

    vec = v.vector().get_local()
    for i in range(len(vec)):
        Vert = dolfin.MeshEntity(mesh, 0, DofToVert[i])
        globalIndex = Vert.global_index()
        VGlobal[globalIndex] = vec[i]
    VGlobal = SyncSum(VGlobal)

    # Use the inverse mapping to se the DOF values of a boundary
    # function
    surface_space = dolfin.FunctionSpace(b_mesh, "CG", 1)
    surface_function = dolfin.Function(surface_space)
    mapa = b_mesh.entity_map(0)
    DofToVert = dolfin.dof_to_vertex_map(dolfin.FunctionSpace(b_mesh, "CG", 1))

    LocValues = surface_function.vector().get_local()
    for i in range(len(LocValues)):
        VolVert = dolfin.MeshEntity(mesh, 0, mapa[int(DofToVert[i])])
        GlobalIndex = VolVert.global_index()
        LocValues[i] = VGlobal[GlobalIndex]

    surface_function.vector().set_local(LocValues)
    surface_function.vector().apply('')
    return surface_function


def boundary_to_mesh(f, mesh):
    """ Take a CG1 function f defined on a surface mesh and return a
    volume vector with same values on boundary but zero in volume
    """
    b_mesh = f.function_space().mesh()
    SpaceV = dolfin.FunctionSpace(mesh, "CG", 1)
    SpaceB = dolfin.FunctionSpace(b_mesh, "CG", 1)

    F = dolfin.Function(SpaceV)
    GValues = numpy.zeros(F.vector().size())

    map = b_mesh.entity_map(0)  # Vertex map from boundary mesh to parent mesh
    d2v = dolfin.dof_to_vertex_map(SpaceB)
    v2d = dolfin.vertex_to_dof_map(SpaceV)

    dof = SpaceV.dofmap()
    imin, imax = dof.ownership_range()

    for i in range(f.vector().local_size()):
        GVertID = dolfin.Vertex(b_mesh, d2v[i]).index()  # Local Vertex ID for given dof on boundary mesh
        PVertID = map[GVertID]  # Local Vertex ID of parent mesh
        PDof = v2d[PVertID]  # Dof on parent mesh
        value = f.vector()[i]  # Value on local processor
        GValues[dof.local_to_global_index(PDof)] = value
    GValues = SyncSum(GValues)

    F.vector().set_local(GValues[imin:imax])
    F.vector().apply("")
    return F


def vector_boundary_to_mesh(boundary_func, mesh):
    """
    Transfer a Vector-CG1 function from a Boundary mesh to
    a CG-1 function living on the Parent mesh (where all interior
    values are 0). This function is only meant to be called internally in
    pyadjoint, or for verification purposes.
    """
    V = dolfin.VectorFunctionSpace(mesh, "CG", 1)
    vb_split = boundary_func.split(deepcopy=True)
    v_vol = []
    for vb in vb_split:
        v_vol.append(boundary_to_mesh(vb, mesh))
    scalar_to_vec = dolfin.FunctionAssigner(V, [v.function_space()
                                                for v in v_vol])
    v_out = dolfin.Function(V)
    scalar_to_vec.assign(v_out, v_vol)
    return v_out


def vector_mesh_to_boundary(func, b_mesh):
    v_split = func.split(deepcopy=True)
    v_b = []
    for v in v_split:
        v_b.append(mesh_to_boundary(v, b_mesh))
    Vb = dolfin.VectorFunctionSpace(b_mesh, "CG", 1)
    vb_out = dolfin.Function(Vb)
    scalar_to_vec = dolfin.FunctionAssigner(Vb, [v.function_space() for
                                                 v in v_b])
    scalar_to_vec.assign(vb_out, v_b)
    return vb_out


def transfer_from_boundary(*args, **kwargs):
    """
    Transfers values from a CG1 function on the BoundaryMesh to its
    original mesh
    """
    annotate = annotate_tape(kwargs)
    with stop_annotating():
        output = vector_boundary_to_mesh(*args)
    output = create_overloaded_object(output)

    if annotate:
        block = SurfaceTransferBlock(args[0], args[1])
        tape = get_working_tape()
        tape.add_block(block)
        block.add_output(output.block_variable)

    return output


class SurfaceTransferBlock(Block):
    def __init__(self, func, other, ad_block_tag=None):
        super(SurfaceTransferBlock, self).__init__(ad_block_tag=ad_block_tag)
        self.add_dependency(func)
        self.add_dependency(func.function_space().mesh())

    @no_annotations
    def evaluate_adj(self, markings=False):
        adj_input = self.get_outputs()[0].adj_value
        if adj_input is None:
            return
        W = self.get_outputs()[0].output.function_space()
        b_mesh = self.get_dependencies()[1].output
        adj_value = dolfin.Function(W)
        adj_value.vector()[:] = adj_input
        adj_output = vector_mesh_to_boundary(adj_value, b_mesh)
        self.get_dependencies()[0].add_adj_output(adj_output.vector())

    @no_annotations
    def evaluate_tlm(self, markings=False):
        tlm_input = self.get_dependencies()[0].tlm_value
        if tlm_input is None:
            return
        W = self.get_outputs()[0].output.function_space()
        output = vector_boundary_to_mesh(tlm_input, W.mesh())
        self.get_outputs()[0].add_tlm_output(output)

    @no_annotations
    def evaluate_hessian(self, markings=False):
        hessian_input = self.get_outputs()[0].hessian_value
        if hessian_input is None:
            return
        W = self.get_outputs()[0].output.function_space()
        mesh = self.get_dependencies()[1].output
        hessian_value = dolfin.Function(W)
        hessian_value.vector()[:] = hessian_input
        hessian_output = vector_mesh_to_boundary(hessian_value, mesh)
        self.get_dependencies()[0].add_hessian_output(hessian_output.vector())

    @no_annotations
    def recompute(self):
        deps = self.get_dependencies()
        W = self.get_outputs()[0].output.function_space()
        deps[0].saved_output.set_allow_extrapolation(True)
        output = vector_boundary_to_mesh(deps[0].saved_output, W.mesh())
        self.get_outputs()[0].checkpoint = output


def transfer_to_boundary(*args, **kwargs):
    """
    Transfers values from a CG1 function on a mesh to its corresponding
    BoundaryMesh.
    """
    annotate = annotate_tape(kwargs)
    with stop_annotating():
        output = vector_mesh_to_boundary(*args)
    output = create_overloaded_object(output)

    if annotate:
        block = VolumeTransferBlock(args[0], args[1])
        tape = get_working_tape()
        tape.add_block(block)
        block.add_output(output.block_variable)

    return output


class VolumeTransferBlock(Block):
    def __init__(self, func, other, ad_block_tag=None):
        super(VolumeTransferBlock, self).__init__(ad_block_tag=ad_block_tag)
        self.add_dependency(func)
        self.add_dependency(func.function_space().mesh())

    @no_annotations
    def evaluate_adj(self, markings=False):
        adj_input = self.get_outputs()[0].adj_value
        if adj_input is None:
            return
        W = self.get_outputs()[0].output.function_space()
        b_mesh = self.get_dependencies()[1].output
        adj_value = dolfin.Function(W, adj_input)
        adj_output = vector_boundary_to_mesh(adj_value, b_mesh)
        self.get_dependencies()[0].add_adj_output(adj_output.vector())

    @no_annotations
    def evaluate_tlm(self, markings=False):
        tlm_input = self.get_dependencies()[0].tlm_value
        if tlm_input is None:
            return
        W = self.get_outputs()[0].output.function_space()
        output = vector_mesh_to_boundary(tlm_input, W.mesh())
        self.get_outputs()[0].add_tlm_output(output)

    @no_annotations
    def evaluate_hessian(self, markings=False):
        hessian_input = self.get_outputs()[0].hessian_value
        if hessian_input is None:
            return
        W = self.get_outputs()[0].output.function_space()
        mesh = self.get_dependencies()[1].output
        hessian_value = dolfin.Function(W, hessian_input)
        hessian_output = vector_boundary_to_mesh(hessian_value, mesh)
        self.get_dependencies()[0].add_hessian_output(hessian_output.vector())

    @no_annotations
    def recompute(self):
        deps = self.get_dependencies()
        W = self.get_outputs()[0].output.function_space()
        deps[0].saved_output.set_allow_extrapolation(True)
        output = vector_mesh_to_boundary(deps[0].saved_output, W.mesh())
        self.get_outputs()[0].checkpoint = output



import dolfin


from pyadjoint.tape import no_annotations
from pyadjoint.overloaded_type import FloatingType
from fenics_adjoint.blocks import DirichletBCBlock

# TODO: Might need/want some way of creating a new DirichletBCBlock if DirichletBC is assigned
#       new boundary values/function.


class DirichletBC(FloatingType, dolfin.DirichletBC):
    def __init__(self, *args, **kwargs):
        super(DirichletBC, self).__init__(*args, **kwargs)

        FloatingType.__init__(self,
                              *args,
                              block_class=DirichletBCBlock,
                              _ad_args=args,
                              _ad_kwargs=kwargs,
                              _ad_floating_active=True,
                              annotate=kwargs.pop("annotate", True),
                              ad_block_tag=kwargs.pop("ad_block_tag", None),
                              **kwargs)

        # Call backend constructor after popped AD specific keyword args.
        dolfin.DirichletBC.__init__(self, *args, **kwargs)

    @no_annotations
    def apply(self, *args, **kwargs):
        for arg in args:
            if not hasattr(arg, "bcs"):
                arg.bcs = []
            arg.bcs.append(self)
        return dolfin.DirichletBC.apply(self, *args, **kwargs)

    def _ad_create_checkpoint(self):
        deps = self.block.get_dependencies()
        if len(deps) <= 0:
            # We don't have any dependencies so the supplied value was not an OverloadedType.
            # Most probably it was just a float that is immutable so will never change.
            return None

        return deps[0]

    def _ad_restore_at_checkpoint(self, checkpoint):
        if checkpoint is not None:
            self.set_value(checkpoint.saved_output)
        return self



import dolfin

from pyadjoint.adjfloat import AdjFloat
from pyadjoint.block import Block
from pyadjoint.overloaded_type import OverloadedType, FloatingType
from pyadjoint.tape import annotate_tape

_REMEMBERED_EXPRESSION_ATTRIBUTES = ["block_variable", "block", "block_class", "annotate_tape",
                                     "user_defined_derivatives", "ad_ignored_attributes", "_ad_attributes_dict"]
_IGNORED_EXPRESSION_ATTRIBUTES = ['_ad_init_object', '_cppcode', '_parameters', '_user_parameters', 'degree',
                                  '_cpp_object', '_ad_dim', '__init_subclass__', '_ufl_required_properties_',
                                  '__radd__', '__rsub__', 'ufl_evaluate', 'rename', 'geometric_dimension',
                                  'ufl_enable_profiling', '_ufl_profiling__del__', '_ad_outputs',
                                  '__getnewargs__', '_ufl_is_index_free_', '_ufl_num_typecodes_',
                                  '_ufl_is_restriction_', '_repr_png_', '__getitem__', '_ufl_noslots_',
                                  '_ufl_num_ops_',
                                  '__subclasshook__', '__pos__', '__sub__', 'ufl_domain', '__getattribute__', '_repr',
                                  '__weakref__', '_ufl_is_scalar_', '__delattr__', '__sizeof__', 'evaluate',
                                  '_ufl_err_str_', '__module__', 'get_generic_function', '__rdiv__',
                                  '_ufl_is_shaping_',
                                  '__init__', 'update', 'tlm_value', '_ad_restore_at_checkpoint', '_ad_mul',
                                  '__format__', '__swig_destroy__', '_ad_annotate_block', '_ad_will_add_as_output',
                                  '_ufl_language_operators_', '_ufl_shape', '_ad_add', 'get_property', '__ne__',
                                  'block_class', 'user_parameters', '__call__', '__reduce_ex__', '__xor__', 'id',
                                  '_ad_attributes_dict', 'ufl_shape', '__str__', 'annotate_tape', 'value_dimension',
                                  '__bool__', 'output_block_class', '_count', '__round__', '__abs__',
                                  '_ad_output_kwargs', 'restrict', '_globalcount', '_ad_to_list', '__iter__',
                                  '__disown__', '_ufl_compute_hash_', '_ad_dot', '_ufl_function_space',
                                  'ufl_disable_profiling', '_ad_function_space', 'eval_cell',
                                  '_ad_annotate_output_block', 'ufl_free_indices', '_ufl_profiling__init__', '__add__',
                                  '_ufl_is_differential_', 'create_block_variable', '_ad_assign_numpy', '__ge__',
                                  '_value_shape', '_ad_output_args', '__dict__', '_ufl_handler_name_', '__reduce__',
                                  'dx', 'value_shape', 'eval', 'stop_floating', '_ufl_required_methods_',
                                  'is_cellwise_constant', '__pow__', 'this', '__doc__', 'str', '_ad_initialized',
                                  'user_defined_derivatives', 'thisown', '_ufl_is_abstract_', '__len__', '__slots__',
                                  '__class__', '_ufl_is_terminal_', '__rpow__', '__lt__', '__div__', '__del__',
                                  '__unicode__', '_ufl_obj_init_counts_', 'value_size', '_ad_kwargs', '__repr__',
                                  '_ufl_obj_del_counts_', '_ufl_evaluate_scalar_', '__neg__', 'ufl_domains', 'count',
                                  'cppcode', '__eq__', '_ad_convert_type', 'ad_ignored_attributes', 'set_property',
                                  'label', '_repr_latex_', 'name', '__le__', '__nonzero__', '_ad_copy',
                                  '_ad_floating_active', 'output_block', '_ufl_is_evaluation_',
                                  'original_block_variable', '__truediv__', 'adj_value', '__float__', '__hash__',
                                  '_ufl_typecode_', '_ufl_all_classes_', '_ufl_is_terminal_modifier_', 'T',
                                  'compute_vertex_values', 'ufl_index_dimensions', '__floordiv__', '__dir__',
                                  'value_rank', 'set_generic_function', 'ufl_function_space',
                                  '_ufl_terminal_modifiers_', '__new__', '_ad_create_checkpoint', '_ad_args',
                                  '_ufl_signature_data_', '_ad_will_add_as_dependency', 'block', 'block_variable',
                                  '_hash', '_ufl_is_literal_', '_function_space', '_ufl_coerce_', 'ufl_element',
                                  'parameters', '__setattr__', '_ufl_class_', '_ufl_is_in_reference_frame_',
                                  '_ufl_expr_reconstruct_', '_ufl_all_handler_names_', '_ufl_regular__init__',
                                  '__rmul__', '__rtruediv__', '__gt__', '_ufl_regular__del__', 'ufl_operands',
                                  '__mul__']


class BaseExpression(FloatingType):
    def __init__(self, *args, **kwargs):
        annotate = annotate_tape(kwargs)
        self._ad_initialized = False
        self._ad_attributes_dict = {}
        self.ad_ignored_attributes = []
        self.user_defined_derivatives = {}
        FloatingType.__init__(self, *args,
                              block_class=ExpressionBlock,
                              annotate=annotate,
                              _ad_args=[self],
                              _ad_floating_active=True,
                              **kwargs)
        self._ad_initialized = True
        self._cached_fs = {}

    def _ad_function_space(self, mesh):
        if mesh not in self._cached_fs:
            element = self.ufl_element()
            fs_element = element.reconstruct(cell=mesh.ufl_cell())
            self._cached_fs[mesh] = dolfin.FunctionSpace(mesh, fs_element)
        return self._cached_fs[mesh]

    def save_attribute(self, key, value):
        self._ad_attributes_dict[key] = value

    def _ad_create_checkpoint(self):
        ret = {}
        for k in self._ad_attributes_dict:
            v = self._ad_attributes_dict[k]
            if isinstance(v, OverloadedType):
                ret[k] = v.block_variable.saved_output
            else:
                ret[k] = v
        return ret

    def _ad_restore_at_checkpoint(self, checkpoint):
        for k in checkpoint:
            self._ad_attributes_dict[k] = checkpoint[k]
            self._ad_backend_setattr_(k, checkpoint[k])
        return self

    def _ad_backend_setattr_(self, k, v):
        object.__setattr__(self, k, v)

    def __setattr__(self, k, v):
        if k == "_ad_initialized" or self._ad_initialized is False:
            object.__setattr__(self, k, v)
        elif k in _REMEMBERED_EXPRESSION_ATTRIBUTES:
            object.__setattr__(self, k, v)
        else:
            self._ad_backend_setattr_(k, v)
            if k not in _IGNORED_EXPRESSION_ATTRIBUTES:
                self.save_attribute(k, v)


class CompiledExpression(BaseExpression, dolfin.CompiledExpression):
    def __init__(self, *args, **kwargs):
        annotate = annotate_tape(kwargs)
        BaseExpression.__init__(self, *args, **kwargs, annotate=annotate)
        kwargs_copy = kwargs.copy()
        for k in kwargs:
            v = kwargs[k]
            if isinstance(v, OverloadedType):
                if hasattr(v, "_cpp_object"):
                    kwargs[k] = v._cpp_object

        dolfin.CompiledExpression.__init__(self, *args, **kwargs)

        for k, val in kwargs_copy.items():
            if hasattr(self._cpp_object, k):
                self.save_attribute(k, val)

    def _ad_backend_setattr_(self, k, v):
        if isinstance(v, OverloadedType):
            if hasattr(v, "_cpp_object"):
                v = v._cpp_object
        dolfin.CompiledExpression.__setattr__(self, k, v)


class UserExpression(BaseExpression, dolfin.UserExpression):
    def __init__(self, *args, **kwargs):
        annotate = annotate_tape(kwargs)
        BaseExpression.__init__(self, *args, **kwargs, annotate=annotate)
        dolfin.UserExpression.__init__(self, *args, **kwargs)

        for k, v in kwargs.items():
            self.save_attribute(k, v)

    def _ad_backend_setattr_(self, k, v):
        dolfin.UserExpression.__setattr__(self, k, v)


class Expression(BaseExpression, dolfin.Expression):
    def __init__(self, *args, **kwargs):
        annotate = annotate_tape(kwargs)
        BaseExpression.__init__(self, *args, **kwargs, annotate=annotate)
        dolfin.Expression.__init__(self, *args, **kwargs)

        for k, v in kwargs.items():
            if k not in _IGNORED_EXPRESSION_ATTRIBUTES:
                self.save_attribute(k, v)

    def _ad_backend_setattr_(self, k, v):
        dolfin.Expression.__setattr__(self, k, v)


class ExpressionBlock(Block):
    def __init__(self, expression):
        super(ExpressionBlock, self).__init__()
        self.expression = expression
        self.dependency_keys = {}

        for key in expression._ad_attributes_dict:
            parameter = expression._ad_attributes_dict[key]
            if isinstance(parameter, OverloadedType):
                self.add_dependency(parameter)
                self.dependency_keys[parameter] = key

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        adj_inputs = adj_inputs[0]
        c = block_variable.output
        if c not in self.expression.user_defined_derivatives:
            return None

        for key in self.expression._ad_attributes_dict:
            if key not in self.expression.ad_ignored_attributes:
                setattr(self.expression.user_defined_derivatives[c], key,
                        self.expression._ad_attributes_dict[key])

        adj_output = None
        for adj_pair in adj_inputs:
            adj_input = adj_pair[0]
            V = adj_pair[1]
            if adj_output is None:
                adj_output = 0.0

            interp = dolfin.interpolate(self.expression.user_defined_derivatives[c], V)
            if isinstance(c, (dolfin.Constant, AdjFloat)):
                adj_output += adj_input.inner(interp.vector())
            else:
                vec = adj_input * interp.vector()
                adj_func = dolfin.Function(V, vec)

                num_sub_spaces = V.num_sub_spaces()
                if num_sub_spaces > 1:
                    for i in range(num_sub_spaces):
                        adj_output += dolfin.interpolate(adj_func.sub(i), c.function_space()).vector()
                else:
                    adj_output += dolfin.interpolate(adj_func, c.function_space()).vector()
        return adj_output

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        # Restore _ad_attributes_dict.
        block_variable.saved_output

        tlm_output = 0.
        tlm_used = False
        for block_variable in self.get_dependencies():
            tlm_input = block_variable.tlm_value
            if tlm_input is None:
                continue

            c = block_variable.output
            if c not in self.expression.user_defined_derivatives:
                continue

            for key in self.expression._ad_attributes_dict:
                if key not in self.expression.ad_ignored_attributes:
                    setattr(self.expression.user_defined_derivatives[c], key, self.expression._ad_attributes_dict[key])

            tlm_used = True
            tlm_output += tlm_input * self.expression.user_defined_derivatives[c]
        if not tlm_used:
            return None
        return tlm_output

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        hessian_inputs = hessian_inputs[0]
        adj_inputs = adj_inputs[0]
        c1 = block_variable.output

        if c1 not in self.expression.user_defined_derivatives:
            return None

        first_deriv = self.expression.user_defined_derivatives[c1]
        for key in self.expression._ad_attributes_dict:
            if key not in self.expression.ad_ignored_attributes:
                setattr(first_deriv, key, self.expression._ad_attributes_dict[key])

        hessian_output = None
        for _, bo2 in relevant_dependencies:
            c2 = bo2.output
            tlm_input = bo2.tlm_value

            if tlm_input is None:
                continue

            if c2 not in first_deriv.user_defined_derivatives:
                continue

            second_deriv = first_deriv.user_defined_derivatives[c2]
            for key in self.expression._ad_attributes_dict:
                if key not in self.expression.ad_ignored_attributes:
                    setattr(second_deriv, key, self.expression._ad_attributes_dict[key])

            for adj_pair in adj_inputs:
                adj_input = adj_pair[0]
                V = adj_pair[1]

                if hessian_output is None:
                    hessian_output = 0.0

                # TODO: Seems we can only project and not interpolate ufl.algebra.Product in dolfin.
                #       Consider the difference and which actually makes sense here.
                interp = dolfin.project(tlm_input * second_deriv, V)
                if isinstance(c1, (dolfin.Constant, AdjFloat)):
                    hessian_output += adj_input.inner(interp.vector())
                else:
                    vec = adj_input * interp.vector()
                    hessian_func = dolfin.Function(V, vec)

                    num_sub_spaces = V.num_sub_spaces()
                    if num_sub_spaces > 1:
                        for i in range(num_sub_spaces):
                            hessian_output += dolfin.interpolate(hessian_func.sub(i), c1.function_space()).vector()
                    else:
                        hessian_output += dolfin.interpolate(hessian_func, c1.function_space()).vector()

        for hessian_pair in hessian_inputs:
            if hessian_output is None:
                hessian_output = 0.0
            hessian_input = hessian_pair[0]
            V = hessian_pair[1]

            interp = dolfin.interpolate(first_deriv, V)
            if isinstance(c1, (dolfin.Constant, AdjFloat)):
                hessian_output += hessian_input.inner(interp.vector())
            else:
                vec = hessian_input * interp.vector()
                hessian_func = dolfin.Function(V, vec)

                num_sub_spaces = V.num_sub_spaces()
                if num_sub_spaces > 1:
                    for i in range(num_sub_spaces):
                        hessian_output += dolfin.interpolate(hessian_func.sub(i), c1.function_space()).vector()
                else:
                    hessian_output += dolfin.interpolate(hessian_func, c1.function_space()).vector()
        return hessian_output

    def recompute_component(self, inputs, block_variable, idx, prepared):
        checkpoint = self.get_outputs()[0].checkpoint

        if checkpoint:
            for bv in self.get_dependencies():
                key = self.dependency_keys[bv.output]
                checkpoint[key] = bv.saved_output

    def __str__(self):
        return "Expression block"



import dolfin
from pyadjoint.overloaded_type import create_overloaded_object


def interpolate(*args, **kwargs):
    """Interpolation is overloaded to ensure that the returned Function object is overloaded.
    We are not able to annotate the interpolation call at the moment.

    """
    kwargs.pop("ad_block_tag", None)
    output = dolfin.interpolate(*args, **kwargs)
    return create_overloaded_object(output)



import dolfin
import ufl_legacy as ufl

from pyadjoint.tape import get_working_tape, stop_annotating, annotate_tape
from .blocks import SolveVarFormBlock, SolveLinearSystemBlock

# Type dependencies

# TODO: Clean up: some inaccurate comments. Reused code. Confusing naming with dFdm when denoting the control as c.


def solve(*args, **kwargs):
    """This solve routine wraps the real Dolfin solve call. Its purpose is to annotate the model,
    recording what solves occur and what forms are involved, so that the adjoint and tangent linear models may be
    constructed automatically by pyadjoint.

    To disable the annotation, just pass :py:data:`annotate=False` to this routine, and it acts exactly like the
    Dolfin solve call. This is useful in cases where the solve is known to be irrelevant or diagnostic
    for the purposes of the adjoint computation (such as projecting fields to other function spaces
    for the purposes of visualisation).

    The overloaded solve takes optional callback functions to extract adjoint solutions.
    All of the callback functions follow the same signature, taking a single argument of type Function.

    Keyword Args:
        adj_cb (function, optional): callback function supplying the adjoint solution in the interior.
            The boundary values are zero.
        adj_bdy_cb (function, optional): callback function supplying the adjoint solution on the boundary.
            The interior values are not guaranteed to be zero.
        adj2_cb (function, optional): callback function supplying the second-order adjoint solution in the interior.
            The boundary values are zero.
        adj2_bdy_cb (function, optional): callback function supplying the second-order adjoint solution on
            the boundary. The interior values are not guaranteed to be zero.

    """
    ad_block_tag = kwargs.pop("ad_block_tag", None)
    annotate = annotate_tape(kwargs)
    if annotate:
        tape = get_working_tape()

        solve_block_type = SolveVarFormBlock
        if not isinstance(args[0], ufl.equation.Equation):
            solve_block_type = SolveLinearSystemBlock

        sb_kwargs = solve_block_type.pop_kwargs(kwargs)
        sb_kwargs.update(kwargs)
        block = solve_block_type(*args, ad_block_tag=ad_block_tag, **sb_kwargs)
        tape.add_block(block)

    with stop_annotating():
        output = dolfin.solve(*args, **kwargs)

    if annotate:
        if hasattr(args[1], "create_block_variable"):
            block_variable = args[1].create_block_variable()
        else:
            block_variable = args[1].function.create_block_variable()
        block.add_output(block_variable)

    return output



import dolfin

backend_fs_sub = dolfin.FunctionSpace.sub


def _fs_sub(self, i):
    V = backend_fs_sub(self, i)
    V._ad_parent_space = self
    return V


dolfin.FunctionSpace.sub = _fs_sub

dolfin.backend_fs_collapse = dolfin.FunctionSpace.collapse


def _fs_collapse(self, collapsed_dofs=False):
    """Overloaded FunctionSpace.collapse to limit the amount of MPI communicator created.
    """
    if not hasattr(self, "_ad_collapsed_space"):
        # Create collapsed space
        self._ad_collapsed_space = dolfin.backend_fs_collapse(self, collapsed_dofs=True)

    if collapsed_dofs:
        return self._ad_collapsed_space
    else:
        return self._ad_collapsed_space[0]


dolfin.FunctionSpace.collapse = _fs_collapse



import dolfin

from fenics_adjoint.utils import function_from_vector


from . import SolveLinearSystemBlock
from .assembly import assemble_adjoint_value


class LUSolveBlockHelper(object):
    def __init__(self):
        self.forward_solver = None
        self.adjoint_solver = None

    def reset(self):
        self.forward_solver = None
        self.adjoint_solver = None


class LUSolveBlock(SolveLinearSystemBlock):
    def __init__(self, A, u, b, *args, **kwargs):
        super(LUSolveBlock, self).__init__(A, u, b, **kwargs)
        self.lu_solver_parameters = kwargs.pop("lu_solver_parameters")
        self.block_helper = kwargs.pop("block_helper")
        self.method = kwargs.pop("lu_solver_method")

    def _assemble_and_solve_adj_eq(self, dFdu_adj_form, dJdu, compute_bdy):
        dJdu_copy = dJdu.copy()
        bcs = self._homogenize_bcs()

        solver = self.block_helper.adjoint_solver
        if solver is None:
            if self.assemble_system:
                rhs_bcs_form = dolfin.inner(dolfin.Function(self.function_space),
                                            dFdu_adj_form.arguments()[0]) * dolfin.dx
                A, _ = dolfin.assemble_system(dFdu_adj_form, rhs_bcs_form, bcs, **self.assemble_kwargs)
            else:
                A = assemble_adjoint_value(dFdu_adj_form, **self.assemble_kwargs)
                [bc.apply(A) for bc in bcs]
            if self.ident_zeros_tol is not None:
                A.ident_zeros(self.ident_zeros_tol)
            solver = dolfin.LUSolver(A, self.method)
            self.block_helper.adjoint_solver = solver

        solver.parameters.update(self.lu_solver_parameters)
        [bc.apply(dJdu) for bc in bcs]

        adj_sol = dolfin.Function(self.function_space)
        solver.solve(adj_sol.vector(), dJdu)

        adj_sol_bdy = None
        if compute_bdy:
            adj_sol_bdy = function_from_vector(self.function_space, dJdu_copy - assemble_adjoint_value(
                dolfin.action(dFdu_adj_form, adj_sol)))

        return adj_sol, adj_sol_bdy

    def _forward_solve(self, lhs, rhs, func, bcs, **kwargs):
        solver = self.block_helper.forward_solver
        if solver is None:
            if self.assemble_system:
                A, _ = dolfin.assemble_system(lhs, rhs, bcs, **self.assemble_kwargs)
            else:
                A = assemble_adjoint_value(lhs, **self.assemble_kwargs)
                [bc.apply(A) for bc in bcs]

            solver = dolfin.LUSolver(A, self.method)
            self.block_helper.forward_solver = solver

        if self.assemble_system:
            system_assembler = dolfin.SystemAssembler(lhs, rhs, bcs)
            b = dolfin.Function(self.function_space).vector()
            system_assembler.assemble(b)
        else:
            b = assemble_adjoint_value(rhs)
            [bc.apply(b) for bc in bcs]

        if self.ident_zeros_tol is not None:
            A.ident_zeros(self.ident_zeros_tol)

        solver.parameters.update(self.lu_solver_parameters)
        solver.solve(func.vector(), b)
        return func



import dolfin
import numpy
import ufl_legacy as ufl
from pyadjoint import AdjFloat, Block, OverloadedType
from pyadjoint.enlisting import Enlist
from pyadjoint.overloaded_type import (FloatingType, create_overloaded_object,
                                       get_overloaded_class,
                                       register_overloaded_type)
from pyadjoint.tape import (annotate_tape, get_working_tape, no_annotations,
                            stop_annotating)
from ufl_legacy.corealg.traversal import traverse_unique_terminals
from ufl_legacy.formatting.ufl2unicode import ufl2unicode

from fenics_adjoint.blocks import (FunctionEvalBlock, FunctionMergeBlock,
                                   FunctionSplitBlock)
from fenics_adjoint.utils import function_from_vector, gather, linalg_solve, create_function
from .constant import create_constant


def type_cast_function(obj, cls):
    """Type casts Function object `obj` to an instance of `cls`.

    Useful when converting backend.Function to overloaded Function.
    """
    return cls(obj.function_space(), obj._cpp_object)


@register_overloaded_type
class Function(FloatingType, dolfin.Function):
    def __init__(self, *args, **kwargs):
        super(Function, self).__init__(*args,
                                       block_class=kwargs.pop("block_class",
                                                              None),
                                       _ad_floating_active=kwargs.pop(
                                           "_ad_floating_active", False),
                                       _ad_args=kwargs.pop("_ad_args", None),
                                       output_block_class=kwargs.pop(
                                           "output_block_class", None),
                                       _ad_output_args=kwargs.pop(
                                           "_ad_output_args", None),
                                       _ad_outputs=kwargs.pop("_ad_outputs",
                                                              None),
                                       annotate=kwargs.pop("annotate", True),
                                       **kwargs)
        dolfin.Function.__init__(self, *args, **kwargs)

    @classmethod
    def _ad_init_object(cls, obj):
        return type_cast_function(obj, cls)

    def copy(self, *args, **kwargs):
        ad_block_tag = kwargs.pop("ad_block_tag", None)
        annotate = annotate_tape(kwargs)
        c = dolfin.Function.copy(self, *args, **kwargs)
        func = create_overloaded_object(c)

        if annotate:
            if kwargs.pop("deepcopy", False):
                block = FunctionAssignBlock(func, self, ad_block_tag=ad_block_tag)
                tape = get_working_tape()
                tape.add_block(block)
                block.add_output(func.create_block_variable())
            else:
                # TODO: Implement. Here we would need to use floating types.
                pass

        return func

    def assign(self, other, *args, **kwargs):
        """To disable the annotation, just pass :py:data:`annotate=False` to this routine, and it acts exactly like the
        Dolfin assign call."""
        # do not annotate in case of self assignment
        ad_block_tag = kwargs.pop("ad_block_tag", None)
        annotate = annotate_tape(kwargs) and self != other
        if annotate:
            if not isinstance(other, ufl.core.operator.Operator):
                other = create_overloaded_object(other)
            block = FunctionAssignBlock(self, other, ad_block_tag=ad_block_tag)
            tape = get_working_tape()
            tape.add_block(block)

        with stop_annotating():
            ret = super(Function, self).assign(other, *args, **kwargs)

        if annotate:
            block.add_output(self.create_block_variable())

        return ret

    def sub(self, i, deepcopy=False, **kwargs):
        from .function_assigner import FunctionAssigner, FunctionAssignerBlock
        ad_block_tag = kwargs.pop("ad_block_tag", None)
        annotate = annotate_tape(kwargs)
        if deepcopy:
            ret = create_overloaded_object(dolfin.Function.sub(self, i, deepcopy, **kwargs))
            if annotate:
                fa = FunctionAssigner(ret.function_space(), self.function_space())
                block = FunctionAssignerBlock(fa, Enlist(self), ad_block_tag=ad_block_tag)
                tape = get_working_tape()
                tape.add_block(block)
                block.add_output(ret.block_variable)
        else:
            extra_kwargs = {}
            if annotate:
                extra_kwargs = {
                    "block_class": FunctionSplitBlock,
                    "_ad_floating_active": True,
                    "_ad_args": [self, i],
                    "_ad_output_args": [i],
                    "output_block_class": FunctionMergeBlock,
                    "_ad_outputs": [self],
                }
            ret = create_function(self, i, **extra_kwargs)
        return ret

    def split(self, deepcopy=False, **kwargs):
        from .function_assigner import FunctionAssigner, FunctionAssignerBlock
        ad_block_tag = kwargs.pop("ad_block_tag", None)
        annotate = annotate_tape(kwargs)
        num_sub_spaces = dolfin.Function.function_space(self).num_sub_spaces()
        if not annotate:
            if deepcopy:
                ret = tuple(create_overloaded_object(dolfin.Function.sub(self, i, deepcopy, **kwargs))
                            for i in range(num_sub_spaces))
            else:
                ret = tuple(create_function(self, i)
                            for i in range(num_sub_spaces))
        elif deepcopy:
            ret = []
            fs = []
            for i in range(num_sub_spaces):

                f = create_overloaded_object(dolfin.Function.sub(self, i, deepcopy, **kwargs))
                fs.append(f.function_space())
                ret.append(f)
            fa = FunctionAssigner(fs, self.function_space())
            block = FunctionAssignerBlock(fa, Enlist(self), ad_block_tag=ad_block_tag)
            tape = get_working_tape()
            tape.add_block(block)
            for output in ret:
                block.add_output(output.block_variable)
            ret = tuple(ret)
        else:
            ret = tuple(Function(self,
                                 i,
                                 block_class=FunctionSplitBlock,
                                 _ad_floating_active=True,
                                 _ad_args=[self, i],
                                 _ad_output_args=[i],
                                 output_block_class=FunctionMergeBlock,
                                 _ad_outputs=[self])
                        for i in range(num_sub_spaces))
        return ret

    def __call__(self, *args, **kwargs):
        ad_block_tag = kwargs.pop("ad_block_tag", None)
        annotate = False
        if len(args) == 1 and isinstance(args[0], (numpy.ndarray,)):
            annotate = annotate_tape(kwargs)

        if annotate:
            block = FunctionEvalBlock(self, args[0], ad_block_tag=ad_block_tag)
            tape = get_working_tape()
            tape.add_block(block)

        with stop_annotating():
            out = dolfin.Function.__call__(self, *args, **kwargs)

        if annotate:
            out = create_overloaded_object(out)
            block.add_output(out.create_block_variable())

        return out

    def vector(self):
        vec = dolfin.Function.vector(self)
        vec.function = self
        return vec

    @no_annotations
    def _ad_convert_type(self, value, options=None):
        options = {} if options is None else options
        riesz_representation = options.get("riesz_representation", "l2")

        if riesz_representation == "l2":
            return create_overloaded_object(
                function_from_vector(self.function_space(), value)
            )
        elif riesz_representation == "L2":
            ret = Function(self.function_space())
            u = dolfin.TrialFunction(self.function_space())
            v = dolfin.TestFunction(self.function_space())
            M = dolfin.assemble(dolfin.inner(u, v) * dolfin.dx)
            linalg_solve(M, ret.vector(), value)
            return ret
        elif riesz_representation == "H1":
            ret = Function(self.function_space())
            u = dolfin.TrialFunction(self.function_space())
            v = dolfin.TestFunction(self.function_space())
            M = dolfin.assemble(
                dolfin.inner(u, v) * dolfin.dx + dolfin.inner(
                    dolfin.grad(u), dolfin.grad(v)) * dolfin.dx)
            linalg_solve(M, ret.vector(), value)
            return ret
        elif callable(riesz_representation):
            return riesz_representation(value)
        else:
            raise NotImplementedError(
                "Unknown Riesz representation %s" % riesz_representation)

    @no_annotations
    def _ad_create_checkpoint(self):
        if self.block is None:
            # TODO: This might crash if annotate=False, but still using a sub-function.
            #       Because subfunction.copy(deepcopy=True) raises the can't access vector error.
            return self.copy(deepcopy=True)

        dep = self.block.get_dependencies()[0].saved_output
        return dep.sub(self.block.idx, deepcopy=False)

    def _ad_restore_at_checkpoint(self, checkpoint):
        return checkpoint

    @no_annotations
    def _ad_mul(self, other):
        r = get_overloaded_class(dolfin.Function)(self.function_space())
        dolfin.Function.assign(r, self * other)
        return r

    @no_annotations
    def _ad_add(self, other):
        r = get_overloaded_class(dolfin.Function)(self.function_space())
        dolfin.Function.assign(r, self + other)
        return r

    def _ad_dot(self, other, options=None):
        options = {} if options is None else options
        riesz_representation = options.get("riesz_representation", "l2")
        if riesz_representation == "l2":
            return self.vector().inner(other.vector())
        elif riesz_representation == "L2":
            return dolfin.assemble(dolfin.inner(self, other) * dolfin.dx)
        elif riesz_representation == "H1":
            return dolfin.assemble(
                (dolfin.inner(self, other) + dolfin.inner(dolfin.grad(self),
                                                          dolfin.grad(
                    other))) * dolfin.dx)
        else:
            raise NotImplementedError(
                "Unknown Riesz representation %s" % riesz_representation)

    @staticmethod
    def _ad_assign_numpy(dst, src, offset):
        range_begin, range_end = dst.vector().local_range()
        m_a_local = src[offset + range_begin:offset + range_end]
        dst.vector().set_local(m_a_local)
        dst.vector().apply('insert')
        offset += dst.vector().size()
        return dst, offset

    @staticmethod
    def _ad_to_list(m):
        if not hasattr(m, "gather"):
            m_v = m.vector()
        else:
            m_v = m
        m_a = gather(m_v)

        return m_a.tolist()

    def _ad_copy(self):
        r = get_overloaded_class(dolfin.Function)(self.function_space())
        dolfin.Function.assign(r, self)
        return r

    def _ad_dim(self):
        return self.function_space().dim()

    def _ad_imul(self, other):
        vec = self.vector()
        vec *= other

    def _ad_iadd(self, other):
        vec = self.vector()
        # FIXME: PETSc complains when we add the same vector to itself.
        # So we make a copy.
        vec += other.vector().copy()

    def _reduce(self, r, r0):
        vec = self.vector().get_local()
        for i in range(len(vec)):
            r0 = r(vec[i], r0)
        return r0

    def _applyUnary(self, f):
        vec = self.vector()
        npdata = vec.get_local()
        for i in range(len(npdata)):
            npdata[i] = f(npdata[i])
        vec.set_local(npdata)
        vec.apply("insert")

    def _applyBinary(self, f, y):
        vec = self.vector()
        npdata = vec.get_local()
        npdatay = y.vector().get_local()
        for i in range(len(npdata)):
            npdata[i] = f(npdata[i], npdatay[i])
        vec.set_local(npdata)
        vec.apply("insert")

    def __deepcopy__(self, memodict={}):
        return self.copy(deepcopy=True)


class FunctionAssignBlock(Block):
    def __init__(self, func, other, ad_block_tag=None):
        super().__init__(ad_block_tag=ad_block_tag)
        self.other = None
        self.expr = None
        if isinstance(other, OverloadedType):
            self.add_dependency(other, no_duplicates=True)
        elif isinstance(other, float) or isinstance(other, int):
            other = AdjFloat(other)
            self.add_dependency(other, no_duplicates=True)
        elif not (isinstance(other, float) or isinstance(other, int)):
            # Assume that this is a point-wise evaluated UFL expression (firedrake only)
            for op in traverse_unique_terminals(other):
                if isinstance(op, OverloadedType):
                    self.add_dependency(op, no_duplicates=True)
            self.expr = other

    def _replace_with_saved_output(self):
        if self.expr is None:
            return None

        replace_map = {}
        for dep in self.get_dependencies():
            replace_map[dep.output] = dep.saved_output
        return ufl.replace(self.expr, replace_map)

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        V = self.get_outputs()[0].output.function_space()
        adj_input_func = function_from_vector(V, adj_inputs[0])

        if self.expr is None:
            return adj_input_func

        expr = self._replace_with_saved_output()
        return expr, adj_input_func

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx,
                               prepared=None):
        if self.expr is None:
            if isinstance(block_variable.output, AdjFloat):
                try:
                    # Adjoint of a broadcast is just a sum
                    return adj_inputs[0].sum()
                except AttributeError:
                    # Catch the case where adj_inputs[0] is just a float
                    return adj_inputs[0]
            elif isinstance(block_variable.output, dolfin.Constant):
                R = block_variable.output._ad_function_space(prepared.function_space().mesh())
                return self._adj_assign_constant(prepared, R)
            else:
                adj_output = dolfin.Function(
                    block_variable.output.function_space())
                adj_output.assign(prepared)
                return adj_output.vector()
        else:
            # Linear combination
            expr, adj_input_func = prepared
            adj_output = dolfin.Function(adj_input_func.function_space())
            if not isinstance(block_variable.output, dolfin.Constant):
                diff_expr = ufl.algorithms.expand_derivatives(
                    ufl.derivative(expr, block_variable.saved_output, adj_input_func)
                )
                adj_output.assign(diff_expr)
            else:
                mesh = adj_output.function_space().mesh()
                diff_expr = ufl.algorithms.expand_derivatives(
                    ufl.derivative(
                        expr,
                        block_variable.saved_output,
                        create_constant(1., domain=mesh)
                    )
                )
                adj_output.assign(diff_expr)
                return adj_output.vector().inner(adj_input_func.vector())

            if isinstance(block_variable.output, dolfin.Constant):
                R = block_variable.output._ad_function_space(adj_output.function_space().mesh())
                return self._adj_assign_constant(adj_output, R)
            else:
                return adj_output.vector()

    def _adj_assign_constant(self, adj_output, constant_fs):
        r = dolfin.Function(constant_fs)
        shape = r.ufl_shape
        if shape == () or shape[0] == 1:
            # Scalar Constant
            r.vector()[:] = adj_output.vector().sum()
        else:
            # We assume the shape of the constant == shape of the output function if not scalar.
            # This assumption is due to FEniCS not supporting products with non-scalar constants in assign.
            values = []
            for i in range(shape[0]):
                values.append(adj_output.sub(i, deepcopy=True).vector().sum())
            r.assign(dolfin.Constant(values))
        return r.vector()

    def prepare_evaluate_tlm(self, inputs, tlm_inputs, relevant_outputs):
        if self.expr is None:
            return None

        return self._replace_with_saved_output()

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx,
                               prepared=None):
        if self.expr is None:
            return tlm_inputs[0]

        expr = prepared
        dudm = dolfin.Function(block_variable.output.function_space())
        dudmi = dolfin.Function(block_variable.output.function_space())
        for dep in self.get_dependencies():
            if dep.tlm_value:
                dudmi.assign(ufl.algorithms.expand_derivatives(
                    ufl.derivative(expr, dep.saved_output,
                                   dep.tlm_value)))
                dudm.vector().axpy(1.0, dudmi.vector())

        return dudm

    def prepare_evaluate_hessian(self, inputs, hessian_inputs, adj_inputs,
                                 relevant_dependencies):
        return self.prepare_evaluate_adj(inputs, hessian_inputs,
                                         relevant_dependencies)

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs,
                                   block_variable, idx,
                                   relevant_dependencies, prepared=None):
        # Current implementation assumes lincom in hessian,
        # otherwise we need second-order derivatives here.
        return self.evaluate_adj_component(inputs, hessian_inputs,
                                           block_variable, idx, prepared)

    def prepare_recompute_component(self, inputs, relevant_outputs):
        if self.expr is None:
            return None
        return self._replace_with_saved_output()

    def recompute_component(self, inputs, block_variable, idx, prepared):
        if self.expr is None:
            prepared = inputs[0]
        output = dolfin.Function(block_variable.output.function_space())
        output.assign(prepared)
        return output

    def __str__(self):
        rhs = self.expr or self.other or self.get_dependencies()[0].output
        if isinstance(rhs, ufl.core.expr.Expr):
            rhs_str = ufl2unicode(rhs)
        else:
            rhs_str = str(rhs)
        return f"assign({rhs_str})"



import dolfin
from pyadjoint.tape import annotate_tape
from pyadjoint import OverloadedType

__all__ = []
__HDF5File_read__ = dolfin.HDF5File.read


def HDF5File_read(self, *args, **kwargs):
    annotate = annotate_tape(kwargs)
    output = __HDF5File_read__(self, *args, **kwargs)

    if annotate:
        func = args[0]
        if isinstance(func, OverloadedType):
            func.create_block_variable()
    return output


dolfin.HDF5File.read = HDF5File_read


__XDMFFile_read__ = dolfin.XDMFFile.read


def XDMFFile_read(self, *args, **kwargs):
    annotate = annotate_tape(kwargs)
    output = __XDMFFile_read__(self, *args, **kwargs)

    if annotate:
        func = args[0]
        if isinstance(func, OverloadedType):
            func.create_block_variable()
    return output


dolfin.XDMFFile.read = XDMFFile_read

__XDMFFile_read_checkpoint__ = dolfin.XDMFFile.read_checkpoint


def XDMFFile_read_checkpoint(self, *args, **kwargs):
    annotate = annotate_tape(kwargs)
    output = __XDMFFile_read_checkpoint__(self, *args, **kwargs)
    if annotate:
        func = args[0]
        if isinstance(func, OverloadedType):
            func.create_block_variable()
    return output


dolfin.XDMFFile.read_checkpoint = XDMFFile_read_checkpoint



from . import SolveVarFormBlock
import dolfin


class ProjectBlock(SolveVarFormBlock):
    def __init__(self, v, V, output, bcs=[], *args, **kwargs):
        mesh = kwargs.pop("mesh", None)
        if mesh is None:
            mesh = V.mesh()
        dx = dolfin.dx(mesh)
        w = dolfin.TestFunction(V)
        Pv = dolfin.TrialFunction(V)
        a = dolfin.inner(w, Pv) * dx
        L = dolfin.inner(w, v) * dx

        # Pop "function" kwarg if present.
        # This relies on the return value of project == function if given.
        kwargs.pop("function", None)

        super(ProjectBlock, self).__init__(a == L, output, bcs, *args, **kwargs)



import dolfin
from pyadjoint.tape import get_working_tape, annotate_tape, stop_annotating
from pyadjoint.overloaded_type import OverloadedType, create_overloaded_object
from pyadjoint.enlisting import Enlist
from fenics_adjoint.blocks import FunctionAssignerBlock


__all__ = ["FunctionAssigner"]


class FunctionAssigner(dolfin.FunctionAssigner):

    def __init__(self, *args, **kwargs):
        super(FunctionAssigner, self).__init__(*args, **kwargs)
        self.input_spaces = Enlist(args[1])
        self.output_spaces = Enlist(args[0])
        self.adj_assigner = dolfin.FunctionAssigner(args[1],
                                                    args[0],
                                                    **kwargs)

    def assign(self, *args, **kwargs):
        annotate = annotate_tape(kwargs)
        outputs = Enlist(args[0])
        inputs = Enlist(args[1])

        if annotate:
            for i, o in enumerate(outputs):
                if not isinstance(o, OverloadedType):
                    outputs[i] = create_overloaded_object(o)

            for j, i in enumerate(outputs):
                if not isinstance(i, OverloadedType):
                    inputs[j] = create_overloaded_object(i)

            block = FunctionAssignerBlock(self, inputs)
            tape = get_working_tape()
            tape.add_block(block)

        with stop_annotating():
            ret = dolfin.FunctionAssigner.assign(self, outputs.delist(), inputs.delist(), **kwargs)

        if annotate:
            for output in outputs:
                block.add_output(output.block_variable)
        return ret



# flake8: noqa

import dolfin

from .constant import Constant
from .dirichletbc import DirichletBC

# Currently not implemented.
from .expression import Expression, UserExpression, CompiledExpression

# Shape AD specific imports for dolfin
from .mesh import *

from .genericmatrix import *
from .genericvector import *
from .io import *

from .as_backend_type import VectorSpaceBasis
from .function_assigner import *
from .function import Function
from .function_space import *

# Import numpy_adjoint to annotate numpy outputs
import numpy_adjoint



import dolfin
import sys
from pyadjoint.tape import get_working_tape, annotate_tape, stop_annotating, no_annotations
from pyadjoint.block import Block
from pyadjoint.overloaded_type import OverloadedType, FloatingType, register_overloaded_type
from ..shapead_transformations import vector_boundary_to_mesh, vector_mesh_to_boundary


overloaded_meshes = ['IntervalMesh', 'UnitIntervalMesh', 'RectangleMesh',
                     'UnitSquareMesh', 'UnitCubeMesh', 'BoxMesh']
__all__ = ['Mesh', 'BoundaryMesh', 'SubMesh'] + overloaded_meshes


@register_overloaded_type
class Mesh(OverloadedType, dolfin.Mesh):
    def __init__(self, *args, **kwargs):
        # Calling constructor
        super(Mesh, self).__init__(*args, **kwargs)
        dolfin.Mesh.__init__(self, *args, **kwargs)

        self._ad_coordinate_space = None

    def _ad_create_checkpoint(self):
        return self.coordinates().copy()

    def _ad_restore_at_checkpoint(self, checkpoint):
        self.coordinates()[:] = checkpoint
        return self

    def _ad_function_space(self):
        if self._ad_coordinate_space is None:
            self._ad_coordinate_space = dolfin.FunctionSpace(self, self.ufl_coordinate_element())
        return self._ad_coordinate_space


@register_overloaded_type
class BoundaryMesh(FloatingType, dolfin.BoundaryMesh):
    def __init__(self, *args, **kwargs):
        # Calling constructor
        super(BoundaryMesh, self).__init__(*args,
                                           block_class=BoundaryMeshBlock,
                                           _ad_args=args,
                                           _ad_floating_active=True,
                                           annotate=kwargs.pop("annotate", True),
                                           **kwargs)
        dolfin.BoundaryMesh.__init__(self, *args, **kwargs)
        self._ad_coordinate_space = None

    def _ad_create_checkpoint(self):
        return self.coordinates().copy()

    def _ad_restore_at_checkpoint(self, checkpoint):
        self.coordinates()[:] = checkpoint
        return self

    def _ad_function_space(self):
        if self._ad_coordinate_space is None:
            self._ad_coordinate_space = dolfin.FunctionSpace(self, self.ufl_coordinate_element())
        return self._ad_coordinate_space


def overloaded_mesh(mesh_class):
    @register_overloaded_type
    class OverloadedMesh(OverloadedType, mesh_class):
        def __init__(self, *args, **kwargs):
            # Calling constructor
            super(OverloadedMesh, self).__init__(*args, **kwargs)
            mesh_class.__init__(self, *args, **kwargs)
            self._ad_coordinate_space = None

        def _ad_create_checkpoint(self):
            return self.coordinates().copy()

        def _ad_restore_at_checkpoint(self, checkpoint):
            self.coordinates()[:] = checkpoint
            return self

        def _ad_function_space(self):
            if self._ad_coordinate_space is None:
                self._ad_coordinate_space = dolfin.FunctionSpace(self, self.ufl_coordinate_element())
            return self._ad_coordinate_space

    return OverloadedMesh


thismod = sys.modules[__name__]
for name in overloaded_meshes:
    setattr(thismod, name, overloaded_mesh(getattr(dolfin, name)))
    mod = getattr(thismod, name)
    mod.__doc__ = getattr(dolfin, name).init.__doc__

SubMesh = overloaded_mesh(dolfin.SubMesh)
SubMesh.__doc__ = dolfin.SubMesh.__doc__


def overloaded_create(mesh_class):
    __ad_create = mesh_class.create

    def create(*args, **kwargs):
        mesh = __ad_create(*args, **kwargs)
        mesh.__class__ = Mesh
        mesh.__init__()

        return mesh
    create.__doc__ = __ad_create.__doc__
    return create


custom_meshes = ["UnitDiscMesh", "SphericalShellMesh", "UnitTriangleMesh", "BoxMesh", "RectangleMesh"]
for name in custom_meshes:
    mesh_type = getattr(dolfin, name)
    mesh_type.create = overloaded_create(mesh_type)


__backend_ALE_move = dolfin.ALE.move


def move(mesh, vector, **kwargs):
    annotate = annotate_tape(kwargs)
    if annotate:
        assert isinstance(mesh, OverloadedType)
        assert isinstance(vector, OverloadedType)
        tape = get_working_tape()
        block = ALEMoveBlock(mesh, vector, **kwargs)
        tape.add_block(block)

    with stop_annotating():
        output = __backend_ALE_move(mesh, vector)
    if annotate:
        block.add_output(mesh.create_block_variable())
    return output


dolfin.ALE.move = move


class ALEMoveBlock(Block):
    def __init__(self, mesh, vector, **kwargs):
        super(ALEMoveBlock, self).__init__(**kwargs)
        self.add_dependency(mesh)
        self.add_dependency(vector)

    @no_annotations
    def evaluate_adj(self, markings=False):
        adj_value = self.get_outputs()[0].adj_value
        if adj_value is None:
            return
        self.get_dependencies()[0].add_adj_output(adj_value.copy())
        self.get_dependencies()[1].add_adj_output(adj_value)

    @no_annotations
    def evaluate_tlm(self, markings=False):
        tlm_output = None

        for i in range(2):
            tlm_input = self.get_dependencies()[i].tlm_value
            if tlm_input is None:
                continue
            if tlm_output is None:
                tlm_output = tlm_input.copy(deepcopy=True)
            else:
                tlm_output.vector().axpy(1, tlm_input.vector())

        if tlm_output is not None:
            self.get_outputs()[0].add_tlm_output(tlm_output)

    @no_annotations
    def evaluate_hessian(self, markings=False):
        hessian_input = self.get_outputs()[0].hessian_value
        if hessian_input is None:
            return
        self.get_dependencies()[0].add_hessian_output(hessian_input.copy())
        self.get_dependencies()[1].add_hessian_output(hessian_input)

    @no_annotations
    def recompute(self, markings=False):
        mesh = self.get_dependencies()[0].saved_output
        vector = self.get_dependencies()[1].saved_output

        dolfin.ALE.move(mesh, vector, annotate=False)

        self.get_outputs()[0].checkpoint = mesh._ad_create_checkpoint()


class BoundaryMeshBlock(Block):
    def __init__(self, mesh, mesh_type, **kwargs):
        super(BoundaryMeshBlock, self).__init__(**kwargs)
        self.add_dependency(mesh)
        self.mesh_type = mesh_type

    @no_annotations
    def evaluate_adj(self, markings=False):
        adj_value = self.get_outputs()[0].adj_value
        if adj_value is None:
            return

        f = dolfin.Function(dolfin.VectorFunctionSpace(self.get_outputs()[0].saved_output, "CG", 1))
        f.vector()[:] = adj_value
        adj_value = vector_boundary_to_mesh(f, self.get_dependencies()[0].saved_output)
        self.get_dependencies()[0].add_adj_output(adj_value.vector())

    @no_annotations
    def evaluate_tlm(self, markings=False):
        tlm_input = self.get_dependencies()[0].tlm_value
        if tlm_input is None:
            return

        tlm_output = vector_mesh_to_boundary(tlm_input, self.get_outputs()[0].saved_output)
        self.get_outputs()[0].add_tlm_output(tlm_output)

    @no_annotations
    def evaluate_hessian(self, markings=False):
        hessian_input = self.get_outputs()[0].hessian_value
        if hessian_input is None:
            return

        f = dolfin.Function(dolfin.VectorFunctionSpace(self.get_outputs()[0].saved_output, "CG", 1))
        f.vector()[:] = hessian_input
        hessian_value = vector_boundary_to_mesh(f, self.get_dependencies()[0].saved_output)
        self.get_dependencies()[0].add_hessian_output(hessian_value.vector())

    @no_annotations
    def recompute(self):
        mesh = self.get_dependencies()[0].saved_output

        b_mesh = BoundaryMesh(mesh, self.mesh_type, annotate=False)

        self.get_outputs()[0].checkpoint = b_mesh._ad_create_checkpoint()



from pyadjoint import Block, OverloadedType
import numpy
import dolfin
from pyadjoint.reduced_functional_numpy import gather
from fenics_adjoint.utils import constant_from_values


class ConstantAssignBlock(Block):
    def __init__(self, other, ad_block_tag=None):
        super(ConstantAssignBlock, self).__init__(ad_block_tag=ad_block_tag)
        self.assigned_float = False
        self.assigned_list = False
        if isinstance(other, OverloadedType):
            self.add_dependency(other)
            self.assigned_float = isinstance(other, float)
        else:
            # Assuming other is supplied as a numpy.ndarray with dtype=object.
            self.assigned_list = True
            self.dependency_to_index = []
            for i, v in enumerate(other.flat):
                if isinstance(v, OverloadedType):
                    self.add_dependency(v)
                    self.dependency_to_index.append(i)
            self.value = other.copy()

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        if self.assigned_list:
            return gather(adj_inputs[0])
        return None

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        adj_output = adj_inputs[0]
        if self.assigned_float:
            # Convert to float
            adj_output = gather(adj_output)
            adj_output = float(adj_output[0])
        elif self.assigned_list:
            adj_output = prepared[self.dependency_to_index[idx]]
        return adj_output

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        values = tlm_inputs[0]
        if self.assigned_list:
            values = numpy.zeros(self.value.shape)
            for i, tlm_input in enumerate(tlm_inputs):
                values.flat[self.dependency_to_index[i]] = tlm_input
        elif isinstance(values, dolfin.Constant):
            values = values.values()
        return constant_from_values(block_variable.output, values)

    def prepare_evaluate_hessian(self, inputs, hessian_inputs, adj_inputs, relevant_dependencies):
        return self.prepare_evaluate_adj(inputs, hessian_inputs, relevant_dependencies)

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        return self.evaluate_adj_component(inputs, hessian_inputs, block_variable, idx, prepared)

    def recompute_component(self, inputs, block_variable, idx, prepared):
        values = inputs[0]
        if self.assigned_list:
            for i, inp in enumerate(inputs):
                self.value[self.dependency_to_index[i]] = inp
            values = self.value
        elif isinstance(values, dolfin.Constant):
            values = values.values()
        return constant_from_values(block_variable.output, values)



import dolfin
from pyadjoint import Block


class FunctionAssignerBlock(Block):
    def __init__(self, assigner, inputs, ad_block_tag=None):
        super(FunctionAssignerBlock, self).__init__(ad_block_tag=ad_block_tag)
        for i in inputs:
            self.add_dependency(i)
        self.assigner = assigner

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        adj_assigner = self.assigner.adj_assigner
        inp_functions = []
        for i in range(len(adj_inputs)):
            f_in = dolfin.Function(self.assigner.output_spaces[i])
            if adj_inputs[i] is not None:
                f_in.vector()[:] = adj_inputs[i]
            inp_functions.append(f_in)
        out_functions = []
        for j in range(len(self.assigner.input_spaces)):
            f_out = dolfin.Function(self.assigner.input_spaces[j])
            out_functions.append(f_out)
        adj_assigner.assign(self.assigner.input_spaces.delist(out_functions),
                            self.assigner.output_spaces.delist(inp_functions))
        return out_functions

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared=None):
        return prepared[idx].vector()

    def prepare_evaluate_tlm(self, inputs, tlm_inputs, relevant_outputs):
        return self.prepare_recompute_component(tlm_inputs, relevant_outputs)

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared=None):
        return prepared[idx]

    def prepare_evaluate_hessian(self, inputs, hessian_inputs, adj_inputs, relevant_dependencies):
        return self.prepare_evaluate_adj(inputs, hessian_inputs, relevant_dependencies)

    def evaluate_hessian_component(self, inputs, hessian_inputs, adj_inputs, block_variable, idx,
                                   relevant_dependencies, prepared=None):
        return self.evaluate_adj_component(inputs, hessian_inputs,
                                           block_variable, idx, prepared)

    def prepare_recompute_component(self, inputs, relevant_outputs):
        out_functions = []
        for output in self.get_outputs():
            out_functions.append(dolfin.Function(output.output.function_space()))
        dolfin.FunctionAssigner.assign(self.assigner,
                                       self.assigner.output_spaces.delist(out_functions),
                                       self.assigner.input_spaces.delist(inputs))
        return out_functions

    def recompute_component(self, inputs, block_variable, idx, out_functions):
        return out_functions[idx]



import numpy
import dolfin


def constant_from_values(constant, values=None):
    """Returns a new Constant with `constant.values()` while preserving `constant.ufl_shape`.

    If the optional argument `values` is provided, then `values` will be the values of the
    new Constant instead, while still preserving the ufl_shape of `constant`.

    Args:
        constant: A constant with the ufl_shape to preserve.
        values (numpy.array): An optional argument to use instead of constant.values().

    Returns:
        Constant: The created Constant of the same type as `constant`.

    """
    values = constant.values() if values is None else values
    return type(constant)(numpy.reshape(values, constant.ufl_shape))


def function_from_vector(V, vector):
    """Create a new Function from a vector.

    :arg V: The function space
    :arg vector: The vector data.
    """
    if isinstance(vector, dolfin.cpp.la.PETScVector)\
            or isinstance(vector, dolfin.cpp.la.Vector):
        pass
    elif not isinstance(vector, dolfin.Vector):
        # If vector is a fenics_adjoint.Function, which does not inherit
        # dolfin.cpp.function.Function with pybind11
        vector = vector._cpp_object
    r = dolfin.Function(V)
    r.vector()[:] = vector
    return r


def extract_subfunction(u, V):
    component = V.component()
    r = u
    for idx in component:
        r = r.sub(int(idx))
    return r


def as_backend_type(A):
    out = dolfin.as_backend_type(A)
    out._ad_original_ref = A
    return out


def create_function(*args, **kwargs):
    """Initialises a fenics_adjoint.Function object and returns it."""
    from fenics_adjoint import Function
    return Function(*args, **kwargs)


MatrixTypes = (dolfin.cpp.la.Matrix, dolfin.cpp.la.GenericMatrix)


def extract_mesh_from_form(form):
    """Takes in a form and extracts a mesh which can be used to construct function spaces.

    Dolfin only accepts backend.cpp.mesh.Mesh types for function spaces, while firedrake use ufl.Mesh.

    Args:
        form (ufl.Form): Form to extract mesh from

    Returns:
        dolfin.Mesh: The extracted mesh

    """
    return form.ufl_domain().ufl_cargo()


def gather(vec):
    import numpy
    if isinstance(vec, dolfin.cpp.function.Function):
        vec = vec.vector()

    if isinstance(vec, dolfin.cpp.la.GenericVector):
        arr = vec.gather(numpy.arange(vec.size(), dtype='I'))
    elif isinstance(vec, list):
        return list(map(gather, vec))
    else:
        arr = vec  # Assume it's a gathered numpy array already

    return arr


def linalg_solve(A, x, b, *args, **kwargs):
    """Linear system solve that has a firedrake compatible interface.

    Throws away kwargs and uses b.vector() as RHS if
    b is not a GenericVector instance.

    """
    if not isinstance(b, dolfin.GenericVector):
        b = b.vector()
    return dolfin.solve(A, x, b, *args)


def create_function(*args, **kwargs):
    """Initialises a fenics_adjoint.Function object and returns it."""
    from fenics_adjoint import Function
    return Function(*args, **kwargs)



import dolfin


from pyadjoint.tape import stop_annotating

from fenics_adjoint.utils import MatrixTypes

_backend_SystemAssembler_assemble = dolfin.SystemAssembler.assemble
_backend_SystemAssembler_init = dolfin.SystemAssembler.__init__


def SystemAssembler_init(self, *args, **kwargs):
    _backend_SystemAssembler_init(self, *args, **kwargs)

    self._A_form = args[0]
    self._b_form = args[1]


def SystemAssembler_assemble(self, *args, **kwargs):
    with stop_annotating():
        out = _backend_SystemAssembler_assemble(self, *args, **kwargs)

    for arg in args:
        if isinstance(arg, dolfin.cpp.la.GenericVector):
            arg.form = self._b_form
            arg.bcs = self._bcs
        elif isinstance(arg, MatrixTypes):
            arg.form = self._A_form
            arg.bcs = self._bcs
            arg.assemble_system = True
        else:
            raise RuntimeError("Argument type not supported: ", type(arg))
    return out


dolfin.SystemAssembler.assemble = SystemAssembler_assemble
dolfin.SystemAssembler.__init__ = SystemAssembler_init



import dolfin

from pyadjoint.tape import annotate_tape, get_working_tape


from fenics_adjoint.blocks import PETScKrylovSolveBlock, PETScKrylovSolveBlockHelper
from fenics_adjoint.utils import MatrixTypes


class PETScKrylovSolver(dolfin.PETScKrylovSolver):
    def __init__(self, *args, **kwargs):
        self.ad_block_tag = kwargs.pop("ad_block_tag", None)
        dolfin.PETScKrylovSolver.__init__(self, *args, **kwargs)

        A = kwargs.pop("A", None)
        method = kwargs.pop("method", "default")
        preconditioner = kwargs.pop("preconditioner", "default")

        next_arg_idx = 0
        if len(args) > 0 and isinstance(args[0], MatrixTypes):
            A = args[0]
            next_arg_idx = 1
        elif len(args) > 1 and isinstance(args[1], MatrixTypes):
            A = args[1]
            next_arg_idx = 2

        if len(args) > next_arg_idx and isinstance(args[next_arg_idx], str):
            method = args[next_arg_idx]
            next_arg_idx += 1
            if len(args) > next_arg_idx and isinstance(args[next_arg_idx], str):
                preconditioner = args[next_arg_idx]

        self.operator = A
        self.pc_operator = None
        self.method = method
        self.preconditioner = preconditioner
        self.solver_parameters = {}
        self.block_helper = PETScKrylovSolveBlockHelper()
        self._ad_nullspace = None
        if hasattr(self.operator, "_ad_nullspace"):
            self._ad_nullspace = self.operator._ad_nullspace

    def set_operator(self, arg0):
        self.operator = arg0
        if hasattr(self.operator, "_ad_nullspace"):
            self._ad_nullspace = self.operator._ad_nullspace
        self.block_helper = PETScKrylovSolveBlockHelper()
        return dolfin.PETScKrylovSolver.set_operator(self, arg0)

    def set_operators(self, arg0, arg1):
        self.operator = arg0
        if hasattr(self.operator, "_ad_nullspace"):
            self._ad_nullspace = self.operator._ad_nullspace

        self.pc_operator = arg1
        self.block_helper = PETScKrylovSolveBlockHelper()
        return dolfin.PETScKrylovSolver.set_operators(self, arg0, arg1)

    def solve(self, *args, **kwargs):
        annotate = annotate_tape(kwargs)

        if annotate:
            if len(args) == 3:
                block_helper = PETScKrylovSolveBlockHelper()
                A = args[0]
                x = args[1]
                b = args[2]
            elif len(args) == 2:
                block_helper = self.block_helper
                A = self.operator
                x = args[0]
                b = args[1]

            u = x.function
            parameters = self.parameters.copy()
            nonzero_initial_guess = parameters["nonzero_initial_guess"] or False
            ksp_options_prefix = dolfin.PETScKrylovSolver.ksp(self).getOptionsPrefix()

            tape = get_working_tape()
            sb_kwargs = PETScKrylovSolveBlock.pop_kwargs(kwargs)
            block = PETScKrylovSolveBlock(A, x, b,
                                          krylov_solver_parameters=parameters,
                                          block_helper=block_helper,
                                          nonzero_initial_guess=nonzero_initial_guess,
                                          pc_operator=self.pc_operator,
                                          krylov_method=self.method,
                                          krylov_preconditioner=self.preconditioner,
                                          ksp_options_prefix=ksp_options_prefix,
                                          _ad_nullspace=self._ad_nullspace,
                                          ad_block_tag=self.ad_block_tag,
                                          **sb_kwargs)
            tape.add_block(block)

        out = dolfin.PETScKrylovSolver.solve(self, *args, **kwargs)

        if annotate:
            block.add_output(u.create_block_variable())

        return out
